"""Admin site for Structures."""

import functools
import statistics
from typing import Optional

from django import forms
from django.conf import settings
from django.contrib import admin
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.db import models
from django.db.models import Prefetch
from django.db.models.functions import Lower
from django.utils.html import format_html
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from app_utils.django import admin_boolean_icon_html
from app_utils.logging import LoggerAddTag

from . import __title__, app_settings, tasks
from .core.notification_types import NotificationType
from .models import (
    FuelAlert,
    FuelAlertConfig,
    GeneratedNotification,
    JumpFuelAlert,
    JumpFuelAlertConfig,
    Notification,
    Owner,
    OwnerCharacter,
    PocoDetails,
    StarbaseDetail,
    Structure,
    StructureItem,
    StructureService,
    StructureTag,
    Webhook,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def lines_sorted_html(items: list) -> str:
    """Format list into HTML with one item per line."""
    return format_html("<br>".join(sorted(items, key=str.casefold)))


class BaseFuelAlertAdmin(admin.ModelAdmin):
    list_display = ("config", "_owner", "structure")
    list_select_related = (
        "config",
        "structure",
        "structure__owner",
        "structure__eve_solar_system",
        "structure__owner__corporation",
    )
    list_filter = (
        ("config", admin.RelatedOnlyFieldListFilter),
        ("structure", admin.RelatedOnlyFieldListFilter),
        ("structure__owner", admin.RelatedOnlyFieldListFilter),
    )
    ordering = ("config", "structure")

    @admin.display(description=_("owner"))
    def _owner(self, obj):
        return obj.structure.owner

    def has_add_permission(self, *args, **kwargs) -> bool:
        return False

    def has_change_permission(self, *args, **kwargs) -> bool:
        return False


if settings.DEBUG:

    @admin.register(FuelAlert)
    class StructureFuelAlertAdmin(BaseFuelAlertAdmin):
        list_display = tuple(BaseFuelAlertAdmin.list_display) + ("hours",)
        ordering = tuple(BaseFuelAlertAdmin.ordering) + ("-hours",)

    @admin.register(JumpFuelAlert)
    class JumpFuelAlertAdmin(BaseFuelAlertAdmin):
        pass


class BaseFuelAlertConfigAdmin(admin.ModelAdmin):
    list_display = (
        "channel_ping_type",
        "_color",
        "is_enabled",
    )
    list_select_related = True
    list_filter = ("is_enabled",)
    actions = ("send_fuel_notifications",)
    fieldsets = (
        (_("Discord"), {"fields": ("channel_ping_type", "color")}),
        (_("General"), {"fields": ("is_enabled",)}),
    )

    @admin.display(ordering="pk", description=_("id"))
    def _id(self, obj):
        return f"#{obj.pk}"

    @admin.display(ordering="color", description=_("color"))
    def _color(self, obj):
        color = Webhook.Color(obj.color)
        return format_html(
            '<span style="color: {};">&#9646;</span>{}',
            color.css_color,
            color.label,
        )

    @admin.action(description=_("Sent fuel notifications for selected configuration"))
    def send_fuel_notifications(self, request, queryset):
        item_count = 0
        for obj in queryset:
            obj.send_new_notifications(force=True)
            item_count += 1
        tasks.send_queued_messages_for_webhooks(Webhook.objects.filter(is_active=True))
        self.message_user(
            request,
            _("Started sending fuel notifications for %d configurations") % item_count,
        )


@admin.register(FuelAlertConfig)
class StructureFuelAlertConfigAdmin(BaseFuelAlertConfigAdmin):
    list_display = (
        "_id",
        "start",
        "end",
        "repeat",
    ) + tuple(BaseFuelAlertConfigAdmin.list_display)
    fieldsets = (
        (
            _("Timing"),
            {
                "description": _(
                    "Timing configuration for sending fuel notifications. "
                    "Note that the first notification will be sent at the exact "
                    "start hour, and the last notification will be sent one repeat "
                    "before the end hour."
                ),
                "fields": ("start", "end", "repeat"),
            },
        ),
    ) + tuple(BaseFuelAlertConfigAdmin.fieldsets)


@admin.register(JumpFuelAlertConfig)
class JumpFuelAlertConfigAdmin(BaseFuelAlertConfigAdmin):
    list_display = (
        "_id",
        "_threshold",
    ) + tuple(BaseFuelAlertConfigAdmin.list_display)
    fieldsets = (
        ("Fuel levels", {"description": ("tbd."), "fields": ("threshold",)}),
    ) + tuple(BaseFuelAlertConfigAdmin.fieldsets)

    @admin.display(ordering="threshold", description=_("threshold"))
    def _threshold(self, obj) -> str:
        return f"{obj.threshold:,}"


class RenderableNotificationFilter(admin.SimpleListFilter):
    title = _("can be send")
    parameter_name = "notification_renderable"

    def lookups(self, request, model_admin):
        return (
            ("yes", _("yes")),
            ("no", _("no")),
        )

    def queryset(self, request, queryset):
        """Return the filtered queryset"""
        if self.value() == "yes":
            return queryset.annotate_can_be_rendered().filter(can_be_rendered_2=True)
        if self.value() == "no":
            return queryset.annotate_can_be_rendered().filter(can_be_rendered_2=False)
        return queryset


class NotificationBaseAdmin(admin.ModelAdmin):
    list_display = (
        "_notification_id",
        "timestamp",
        "owner",
        "notif_type",
        "_webhooks",
        "_structures",
        "_is_sent",
        "_is_timer_added",
    )
    list_filter = (
        ("owner", admin.RelatedOnlyFieldListFilter),
        RenderableNotificationFilter,
        "is_sent",
        "notif_type",
    )
    ordering = ["-timestamp"]
    actions = (
        "mark_as_sent",
        "mark_as_unsent",
        "send_to_configured_webhooks",
        "add_or_remove_timer",
    )

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.prefetch_related(
            "owner__webhooks",
            "structures",
            "structures__eve_solar_system",
            "structures__eve_type",
            "structures__eve_type__eve_group",
            Prefetch(
                "structures",
                queryset=Structure.objects.filter(webhooks__isnull=False),
                to_attr="structures_with_webhooks",
            ),
        ).select_related("owner", "owner__corporation")

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    @admin.display(description=_("notification ID"))
    def _notification_id(self, obj: Notification):
        return obj.notification_id

    @admin.display(description=_("webhooks"))
    def _webhooks(self, obj: Notification):
        if not obj.can_be_rendered:
            return format_html("<i>{}</i>", _("N/A"))
        webhooks_qs = obj.owner.webhooks.all()
        if obj.structures_with_webhooks:
            webhooks_qs = Webhook.objects.none()
            for structure in obj.structures_with_webhooks:
                webhooks_qs |= structure.webhooks.all()
        names = list(
            {
                webhook.name
                for webhook in webhooks_qs
                if obj.notif_type in webhook.notification_types
            }
        )
        if not names:
            return format_html(
                '<b><span style="color: orange">⚠ {}</span></b>', _("Not configured")
            )
        return lines_sorted_html(names)

    @admin.display(description=_("structures"))
    def _structures(self, obj: Notification) -> Optional[str]:
        if obj.is_structure_related:
            structures = [str(structure) for structure in obj.structures.all()]
            return lines_sorted_html(structures) if structures else "?"
        return None

    @admin.display(description=_("is sent"))
    def _is_sent(self, obj: Notification):
        value = obj.is_sent if obj.can_be_rendered else None
        return admin_boolean_icon_html(value)

    @admin.display(description=_("is timer added"))
    def _is_timer_added(self, obj: Notification):
        value = obj.is_timer_added if obj.can_have_timer else None
        return admin_boolean_icon_html(value)

    @admin.action(description=_("Mark selected notifications as sent"))
    def mark_as_sent(self, request, queryset):
        queryset.update(is_sent=True)
        notif_count = queryset.count()
        self.message_user(request, _("%d notifications marked as sent") % notif_count)

    @admin.action(description=_("Mark selected notifications as unsent"))
    def mark_as_unsent(self, request, queryset):
        queryset.update(is_sent=False)
        notif_count = queryset.count()
        self.message_user(request, _("%d notifications marked as unsent") % notif_count)

    @admin.action(description=_("Send selected notifications to configured webhooks"))
    def send_to_configured_webhooks(self, request, queryset):
        notifs_queued = 0
        for obj in queryset:
            obj: Notification
            if obj.can_be_rendered and obj.relevant_webhooks().exists():
                if obj.send_to_configured_webhooks():
                    notifs_queued += 1

        if notifs_queued:
            tasks.send_queued_messages_for_webhooks(
                Webhook.objects.filter(is_active=True)
            )

        self.message_user(
            request,
            _("Sent %(sent_count)d/%(selected_count)d generated messages.")
            % {"sent_count": notifs_queued, "selected_count": queryset.count()},
        )

    @admin.action(description=_("Process selected notifications for timerboard"))
    def add_or_remove_timer(self, request, queryset):
        notifications_count = 0
        ignored_count = 0
        for obj in queryset:
            if obj.add_or_remove_timer():
                notifications_count += 1
            else:
                ignored_count += 1

        first = (
            _("Added timers from %d notifications to timerboard.") % notifications_count
        )
        second = (
            _("Ignored %d notification(s), which has no relation to timers.")
            % ignored_count
        )
        if ignored_count:
            message = format_lazy("{first} {second}", first=first, second=second)
        else:
            message = first
        message = format_lazy(first)
        self.message_user(request, message)


@admin.register(Notification)
class NotificationAdmin(NotificationBaseAdmin):
    pass


@admin.register(GeneratedNotification)
class GeneratedNotificationAdmin(NotificationBaseAdmin):
    pass


class OwnerCharacterAdminInline(admin.TabularInline):
    model = OwnerCharacter

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False


class DisabledCharactersFilter(admin.SimpleListFilter):
    title = _("has disabled characters")
    parameter_name = "has_disabled_characters"

    def lookups(self, request, model_admin):
        return (
            ("yes", _("yes")),
            ("no", _("no")),
        )

    def queryset(self, request, queryset):
        """Return the filtered queryset"""
        if self.value() == "yes":
            return queryset.annotate_characters_count().filter(
                characters_disabled_count__gt=0
            )
        if self.value() == "no":
            return queryset.annotate_characters_count().filter(
                characters_disabled_count=0
            )
        return queryset


@admin.register(Owner)
class OwnerAdmin(admin.ModelAdmin):
    list_display = (
        "_corporation",
        "_alliance",
        "_characters",
        "_is_active",
        "_webhooks",
        "_has_default_pings_enabled",
        "_ping_groups",
        "_is_alliance_main",
        "_is_sync_ok",
    )
    list_filter = (
        "is_active",
        "is_up",
        DisabledCharactersFilter,
        ("corporation__alliance", admin.RelatedOnlyFieldListFilter),
        "has_default_pings_enabled",
        "is_alliance_main",
    )
    ordering = ["corporation__corporation_name"]
    search_fields = ["corporation__corporation_name"]
    actions = (
        "update_all",
        "update_structures",
        "fetch_notifications",
        "deactivate_owners",
        "activate_owners",
        "reset_characters",
    )
    inlines = (OwnerCharacterAdminInline,)
    filter_horizontal = ("ping_groups", "webhooks")
    fieldsets = (
        (
            None,
            {
                "fields": (
                    "corporation",
                    "_structures_count",
                    "webhooks",
                    "is_alliance_main",
                    "are_pocos_public",
                    "has_default_pings_enabled",
                    "ping_groups",
                    "is_included_in_service_status",
                    "is_active",
                )
            },
        ),
        (
            _("Sync Status"),
            {
                "classes": ("collapse",),
                "fields": (
                    "_are_all_syncs_ok",
                    ("_structures_last_update_fresh", "structures_last_update_at"),
                    (
                        "_notifications_last_update_fresh",
                        "notifications_last_update_at",
                        "_avg_turnaround_time",
                    ),
                    ("_forwarding_last_update_fresh", "forwarding_last_update_at"),
                    ("_assets_last_update_fresh", "assets_last_update_at"),
                ),
            },
        ),
    )

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        """only show custom tags in dropdown"""
        if db_field.name == "webhooks":
            kwargs["queryset"] = Webhook.objects.filter(is_active=True)

        return super().formfield_for_manytomany(db_field, request, **kwargs)

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return (
            qs.select_related("corporation", "corporation__alliance")
            .prefetch_related("ping_groups", "webhooks")
            .annotate_characters_count()
        )

    def get_form(self, *args, **kwargs):
        """Add help text to custom field."""
        help_texts = {
            "_avg_turnaround_time": (
                "For last %d | %d | %d notifications"
                % (
                    app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_SHORT,
                    app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_MEDIUM,
                    app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_LONG,
                )
            ),
            "_are_all_syncs_ok": (
                "True when all of the last successful syncs were within grace periods."
            ),
            "_structures_last_update_fresh": (
                "True when last sync within %s minutes."
                % app_settings.STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES
            ),
            "_notifications_last_update_fresh": (
                "True when last sync within %s minutes."
                % app_settings.STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES
            ),
            "_forwarding_last_update_fresh": (
                "True when last sync within %s minutes."
                % app_settings.STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES
            ),
            "_assets_last_update_fresh": (
                "True when last sync within %s minutes."
                % app_settings.STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES
            ),
        }
        kwargs.update({"help_texts": help_texts})
        return super().get_form(*args, **kwargs)

    def get_readonly_fields(self, request, obj=None):
        if obj:  # editing an existing object
            return tuple(self.readonly_fields) + (
                "assets_last_update_at",
                "corporation",
                "forwarding_last_update_at",
                "notifications_last_update_at",
                "structures_last_update_at",
                "_avg_turnaround_time",
                "_are_all_syncs_ok",
                "_structures_last_update_fresh",
                "_notifications_last_update_fresh",
                "_forwarding_last_update_fresh",
                "_assets_last_update_fresh",
                "_structures_count",
            )
        return self.readonly_fields

    def has_add_permission(self, request):
        return False

    @admin.display(ordering="characters_enabled_count", description=_("characters"))
    def _characters(self, obj: Owner) -> str:
        enabled = obj.characters_enabled_count
        disabled = obj.characters_disabled_count
        if not disabled:
            return enabled
        return f"{enabled} ({disabled})"

    @admin.display(description=_("default pings"), boolean=True)
    def _has_default_pings_enabled(self, obj: Owner):
        return obj.has_default_pings_enabled

    @admin.display(description=_("ping groups"))
    def _ping_groups(self, obj: Owner):
        ping_groups = [ping_group.name for ping_group in obj.ping_groups.all()]
        return lines_sorted_html(ping_groups) if ping_groups else None

    @admin.display(
        ordering="corporation__corporation_name", description=_("corporation")
    )
    def _corporation(self, obj: Owner):
        return obj.corporation.corporation_name

    @admin.display(
        ordering="corporation__alliance__alliance_name", description=_("alliance")
    )
    def _alliance(self, obj: Owner):
        if obj.corporation.alliance:
            return obj.corporation.alliance.alliance_name
        return None

    @admin.display(description=_("webhooks"))
    def _webhooks(self, obj: Owner):
        names = [webhook.name for webhook in obj.webhooks.all()]
        if names:
            return lines_sorted_html(names)

        return format_html(
            '<span style="color: red">⚠ {}</span>',
            _(
                "Notifications can not be sent, "
                "because there is no webhook configured for this owner."
            ),
        )

    @admin.display(description=_("active"), boolean=True)
    def _is_active(self, obj: Owner):
        return obj.is_active

    @admin.display(description=_("alliance main"))
    def _is_alliance_main(self, obj: Owner):
        value = True if obj.is_alliance_main else None
        return admin_boolean_icon_html(value)

    @admin.display(description=_("services up"), boolean=True)
    def _is_sync_ok(self, obj: Owner):
        return obj.is_up if obj.is_active else None

    @admin.action(description=_("Activate selected owners"))
    def activate_owners(self, request, queryset):
        queryset.update(is_active=True)
        self.message_user(request, _("Activated %d owners") % queryset.count())

    @admin.action(description=_("Deactivate selected owner"))
    def deactivate_owners(self, request, queryset):
        queryset.update(is_active=False)
        self.message_user(request, _("Deactivated %d owners") % queryset.count())

    @admin.action(description=_("Reset disabled characters for selected owners"))
    def reset_characters(self, request, queryset):
        owner_pks = queryset.values_list("pk", flat=True)
        OwnerCharacter.objects.filter(
            owner__pk__in=list(owner_pks), is_enabled=False
        ).update(is_enabled=True, disabled_reason="", error_count=0)
        self.message_user(
            request, _("Characters have been reset for %d owners") % len(owner_pks)
        )

    @admin.action(description=_("Update all from EVE server for selected owners"))
    def update_all(self, request, queryset):
        for obj in queryset:
            tasks.update_all_for_owner.delay(obj.pk, user_pk=request.user.pk)  # type: ignore

            text = format_lazy(
                "{first} {second}",
                first=_("Started updating structures and notifications for %s.") % obj,
                second=_("You will receive a notification once it is completed."),
            )
            self.message_user(request, text)

    @admin.action(
        description=_("Update structures from EVE server for selected owners")
    )
    def update_structures(self, request, queryset):
        for obj in queryset:
            tasks.update_structures_for_owner.delay(obj.pk, user_pk=request.user.pk)  # type: ignore
            text = format_lazy(
                "{first} {second}",
                first=_("Started updating structures for %s.") % obj,
                second=_("You will receive a notification once it is completed."),
            )
            self.message_user(request, text)

    @admin.action(
        description=_("Fetch notifications from EVE server for selected owners")
    )
    def fetch_notifications(self, request, queryset):
        for obj in queryset:
            tasks.process_notifications_for_owner.delay(obj.pk, user_pk=request.user.pk)  # type: ignore
            text = format_lazy(
                "{first} {second}",
                first=_("Started fetching notifications for %s.") % obj,
                second=_("You will receive a notification once it is completed."),
            )
            self.message_user(request, text)

    @admin.display(description=_("all syncs OK"), boolean=True)
    def _are_all_syncs_ok(self, obj: Owner):
        return obj.are_all_syncs_ok

    @admin.display(description=_("avg. turnaround time"))
    def _avg_turnaround_time(self, obj: Owner) -> str:
        """Average time between timestamp of notifications an when they are received."""

        def my_format(value) -> str:
            return f"{value:,.0f}" if value else "-"

        max_short = app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_SHORT
        max_medium = app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_MEDIUM
        max_long = app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_LONG
        max_valid = app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_MAX_VALID
        notifications = obj.notification_set.filter(created__isnull=False).order_by(
            "-timestamp"
        )
        data = [
            (rec[0] - rec[1]).total_seconds()
            for rec in notifications.values_list("created", "timestamp")
            if (rec[0] - rec[1]).total_seconds() < max_valid
        ]
        short = statistics.mean(data[:max_short]) if len(data) >= max_short else None
        medium = statistics.mean(data[:max_medium]) if len(data) >= max_medium else None
        long = statistics.mean(data[:max_long]) if len(data) >= max_long else None
        return f"{my_format(short)} | {my_format(medium)} | {my_format(long)}"

    @admin.display(description=_("structures update fresh"), boolean=True)
    def _structures_last_update_fresh(self, obj: Owner) -> bool:
        return obj.is_structure_sync_fresh

    @admin.display(description=_("notifications update fresh"), boolean=True)
    def _notifications_last_update_fresh(self, obj: Owner) -> bool:
        return obj.is_notification_sync_fresh

    @admin.display(description=_("forwarding update fresh"), boolean=True)
    def _forwarding_last_update_fresh(self, obj: Owner) -> bool:
        return obj.is_forwarding_sync_fresh

    @admin.display(description=_("assets update fresh"), boolean=True)
    def _assets_last_update_fresh(self, obj: Owner) -> bool:
        return obj.is_assets_sync_fresh

    @admin.display(description=_("structures Count"))
    def _structures_count(self, obj: Owner) -> int:
        return obj.structures.count()


@admin.register(StructureTag)
class StructureTagAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "description",
        "order",
        "style",
        "is_default",
        "is_user_managed",
    )
    list_filter = ("is_default", "style", "is_user_managed")
    readonly_fields = ("is_user_managed",)

    def has_delete_permission(self, request, obj: Optional[StructureTag] = None):
        return not (obj and not obj.is_user_managed)

    def has_change_permission(self, request, obj: Optional[StructureTag] = None):
        return not (obj and not obj.is_user_managed)


class StructureServiceAdminInline(admin.TabularInline):
    model = StructureService

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class StructureItemAdminInline(admin.TabularInline):
    model = StructureItem

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class StructureStarbaseDetailAdminInline(admin.StackedInline):
    model = StarbaseDetail

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class StructurePocoDetailsAdminInline(admin.StackedInline):
    model = PocoDetails

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class OwnerCorporationsFilter(admin.SimpleListFilter):
    """Custom filter to filter on corporations from owners only"""

    title = _("owner corporation")
    parameter_name = "owner_corporation_id__exact"

    def lookups(self, request, model_admin):
        qs = (
            EveCorporationInfo.objects.filter(structure_owner__isnull=False)
            .values("corporation_id", "corporation_name")
            .distinct()
            .order_by(Lower("corporation_name"))
        )
        return tuple(((obj["corporation_id"], obj["corporation_name"]) for obj in qs))

    def queryset(self, request, queryset):
        if self.value() is None:
            return queryset.all()

        return queryset.filter(owner__corporation__corporation_id=self.value())


class OwnerAllianceFilter(admin.SimpleListFilter):
    """Custom filter to filter on alliances from owners only"""

    title = _("owner alliance")
    parameter_name = "owner_alliance_id__exact"

    def lookups(self, request, model_admin):
        qs = (
            EveAllianceInfo.objects.filter(
                evecorporationinfo__structure_owner__isnull=False
            )
            .values("alliance_id", "alliance_name")
            .distinct()
            .order_by(Lower("alliance_name"))
        )
        return tuple(((obj["alliance_id"], obj["alliance_name"]) for obj in qs))

    def queryset(self, request, queryset):
        if self.value() is None:
            return queryset.all()

        return queryset.filter(owner__corporation__alliance__alliance_id=self.value())


class HasWebhooksListFilter(admin.SimpleListFilter):
    title = _("has webhooks")
    parameter_name = "has_webhooks"

    def lookups(self, request, model_admin):
        return (
            ("y", _("yes")),
            ("n", _("no")),
        )

    def queryset(self, request, queryset):
        if self.value() == "y":
            return queryset.filter(webhooks__isnull=False)
        if self.value() == "n":
            return queryset.filter(webhooks__isnull=True)
        return None


@admin.register(Structure)
class StructureAdmin(admin.ModelAdmin):
    show_full_result_count = True
    list_select_related = (
        "owner",
        "owner__corporation",
        "owner__corporation__alliance",
        "eve_solar_system",
        "eve_solar_system__eve_constellation__eve_region",
        "eve_type",
        "eve_type__eve_group",
        "eve_type__eve_group__eve_category",
        "eve_planet",
        "eve_moon",
    )
    search_fields = [
        "id",
        "name",
        "owner__corporation__corporation_name",
        "eve_solar_system__name",
    ]
    ordering = ["name"]
    list_display = (
        "_name",
        "_owner",
        "_location",
        "_type",
        "_power_mode",
        "_tags",
        "_webhooks",
    )
    list_filter = (
        ("eve_type__eve_group__eve_category", admin.RelatedOnlyFieldListFilter),
        OwnerCorporationsFilter,
        OwnerAllianceFilter,
        ("eve_solar_system", admin.RelatedOnlyFieldListFilter),
        (
            "eve_solar_system__eve_constellation__eve_region",
            admin.RelatedOnlyFieldListFilter,
        ),
        ("eve_type", admin.RelatedOnlyFieldListFilter),
        ("eve_type__eve_group", admin.RelatedOnlyFieldListFilter),
        ("tags", admin.RelatedOnlyFieldListFilter),
        HasWebhooksListFilter,
    )
    actions = ("add_default_tags", "remove_user_tags", "update_generated_tags")
    readonly_fields = tuple(
        (
            obj.name
            for obj in Structure._meta.get_fields()
            if isinstance(obj, models.fields.Field)
            and obj.name not in ["tags", "webhooks"]
        )
    )
    fieldsets = (
        (
            None,
            {
                "fields": (
                    "name",
                    "owner",
                    "eve_solar_system",
                    "eve_type",
                    "tags",
                    "webhooks",
                )
            },
        ),
        (
            _("Status"),
            {
                "classes": ("collapse",),
                "fields": (
                    "state",
                    (
                        "state_timer_start",
                        "state_timer_end",
                    ),
                    "unanchors_at",
                    "fuel_expires_at",
                    "last_online_at",
                    "has_fitting",
                    "has_core",
                ),
            },
        ),
        (
            _("Reinforcement"),
            {
                "classes": ("collapse",),
                "fields": (
                    ("reinforce_hour",),
                    ("next_reinforce_hour", "next_reinforce_apply"),
                ),
            },
        ),
        (
            _("Position"),
            {
                "classes": ("collapse",),
                "fields": ("position_x", "position_y", "position_z"),
            },
        ),
        (
            None,
            {"fields": (("id", "last_updated_at"))},
        ),
    )
    filter_horizontal = ("tags", "webhooks")

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        if db_field.name == "tags":
            kwargs["queryset"] = StructureTag.objects.filter(is_user_managed=True)

        return super().formfield_for_manytomany(db_field, request, **kwargs)

    def get_inlines(self, request, obj: Structure):
        if obj.is_upwell_structure:
            return (
                StructureServiceAdminInline,
                StructureItemAdminInline,
            )

        if obj.is_poco:
            return (StructurePocoDetailsAdminInline,)

        if obj.is_starbase:
            return (StructureStarbaseDetailAdminInline,)

        return []

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.prefetch_related("tags", "webhooks")

    def has_add_permission(self, request):
        return False

    @admin.display(ordering="name", description=_("name"))
    def _name(self, obj: Structure) -> str:
        if obj.name:
            return obj.name
        return obj.location_name

    @admin.display(
        ordering="owner__corporation__corporation_name", description=_("owner")
    )
    def _owner(self, obj: Structure) -> str:
        alliance = obj.owner.corporation.alliance
        return format_html(
            "{}<br>{}",
            obj.owner.corporation,
            alliance if alliance else "",
        )

    @admin.display(ordering="eve_solar_system__name", description=_("location"))
    def _location(self, obj: Structure) -> str:
        return format_html(
            "{}<br>{}",
            obj.location_name,
            obj.eve_solar_system.eve_constellation.eve_region,
        )

    @admin.display(ordering="eve_type__name", description=_("type"))
    def _type(self, obj: Structure):
        return format_html("{}<br>{}", obj.eve_type, obj.eve_type.eve_group)

    @admin.display(description=_("power mode"))
    def _power_mode(self, obj: Structure) -> str:
        return obj.get_power_mode_display()

    @admin.display(description=_("tags"))
    def _tags(self, obj: Structure) -> list:
        return sorted([tag.name for tag in obj.tags.all()])

    @admin.display(description=_("webhooks"))
    def _webhooks(self, obj: Structure):
        names = [webhook.name for webhook in obj.webhooks.all()]
        return lines_sorted_html(names) if names else None

    @admin.action(description=_("Add default tags to selected structures"))
    def add_default_tags(self, request, queryset):
        structure_count = 0
        tags = StructureTag.objects.filter(is_default=True)
        for structure in queryset:
            for tag in tags:
                structure.tags.add(tag)
            structure_count += 1
        tags_count = tags.count()
        self.message_user(
            request,
            _("Added %(tags_count)d default tags to %(structure_count)d structures")
            % {"tags_count": tags_count, "structure_count": structure_count},
        )

    @admin.action(description=_("Remove user tags for selected structures"))
    def remove_user_tags(self, request, queryset):
        structure_count = 0
        for structure in queryset:
            for tag in structure.tags.filter(is_user_managed=True):
                structure.tags.remove(tag)
            structure_count += 1
        self.message_user(
            request, _("Removed all user tags from %d structures") % structure_count
        )

    @admin.action(description=_("Update generated tags for selected structures"))
    def update_generated_tags(self, request, queryset):
        structure_count = 0
        for structure in queryset:
            structure.update_generated_tags(recreate_tags=True)
            structure_count += 1
        self.message_user(
            request, _("Updated all generated tags for %d structures") % structure_count
        )


class WebhookAdminForm(forms.ModelForm):
    owners = forms.ModelMultipleChoiceField(
        queryset=Owner.objects.order_by(Lower("corporation__corporation_name")),
        required=False,
        widget=FilteredSelectMultiple(verbose_name=_("Users"), is_stacked=False),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.instance and self.instance.pk:
            self.fields["owners"].initial = self.instance.owners.all()

    def save(self, commit=True):
        webhook: Webhook = super().save(commit=False)

        if commit:
            webhook.save()

        owners = self.cleaned_data["owners"]
        if webhook.pk:
            self._save_m2m_and_users(webhook, owners)
        else:
            self.save_m2m = functools.partial(
                self._save_m2m_and_users, webhook=webhook, owners=owners
            )

        return webhook

    def _save_m2m_and_users(self, webhook, owners):
        """Save m2m relations incl. users."""
        webhook.owners.set(owners)
        self._save_m2m()


@admin.register(Webhook)
class WebhookAdmin(admin.ModelAdmin):
    form = WebhookAdminForm
    list_display = (
        "name",
        "_ping_groups",
        "_owners",
        "is_active",
        "_is_default",
        "_messages_in_queue",
    )
    list_filter = ("is_active",)
    ordering = ["name"]
    save_as = True
    actions = (
        "test_notification",
        "activate",
        "deactivate",
        "purge_messages",
        "send_messages",
    )
    filter_horizontal = ("ping_groups",)
    fieldsets = (
        (
            None,
            {
                "fields": (
                    "name",
                    "url",
                    "notes",
                    "owners",
                    "notification_types",
                    "ping_groups",
                    "is_active",
                    "is_default",
                ),
            },
        ),
        (
            _("Advanced Options"),
            {
                "classes": ("collapse",),
                "fields": (
                    "language_code",
                    "has_default_pings_enabled",
                    "webhook_type",
                ),
            },
        ),
    )

    def get_form(self, *args, **kwargs):
        form = super().get_form(*args, **kwargs)
        form.base_fields["notification_types"].choices = sorted(
            NotificationType.choices_enabled(), key=lambda k: k[1]
        )
        return form

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.prefetch_related(
            "ping_groups",
            "owners",
            "owners__corporation",
            Prefetch(
                "structures",
                queryset=Structure.objects.select_related(
                    "eve_solar_system",
                    "eve_type",
                    "eve_type__eve_group",
                    "owner",
                    "owner__corporation",
                ),
            ),
        )

    @admin.display(description=_("ping groups"))
    def _ping_groups(self, obj: Webhook):
        ping_groups = [ping_group.name for ping_group in obj.ping_groups.all()]
        return lines_sorted_html(ping_groups) if ping_groups else None

    @admin.display(description=_("enabled for owners or structures"))
    def _owners(self, obj: Webhook):
        configurations = [str(owner) for owner in obj.owners.all()]
        configurations += [
            f"{structure.owner}: {structure}" for structure in obj.structures.all()
        ]
        if not configurations:
            return format_html(
                '<b><span style="color: orange;">⚠ {}</span></b>',
                _("Please add this webhook to an owner or structure to enable it."),
            )
        return lines_sorted_html(configurations)

    @admin.display(description=_("is default"))
    def _is_default(self, obj: Webhook):
        value = True if obj.is_default else None
        return admin_boolean_icon_html(value)

    @admin.display(description=_("messages in queue"))
    def _messages_in_queue(self, obj: Webhook):
        return obj.queue_size()

    @admin.action(description=_("Send test notification to selected webhook"))
    def test_notification(self, request, queryset):
        for obj in queryset:
            tasks.send_test_notifications_to_webhook.delay(
                obj.pk, user_pk=request.user.pk
            )  # type: ignore
            text = format_lazy(
                "{first} {second}",
                first=_("Initiated sending test notification to webhook %s.") % obj,
                second=_("You will receive a notification once it is completed."),
            )
            self.message_user(request, text)

    @admin.action(description=_("Activate selected webhook"))
    def activate(self, request, queryset):
        for obj in queryset:
            obj.is_active = True
            obj.save()
            self.message_user(request, _("You have activated webhook %s") % obj)

    @admin.action(description=_("Deactivate selected webhook"))
    def deactivate(self, request, queryset):
        for obj in queryset:
            obj.is_active = False
            obj.save()
            self.message_user(request, _("You have de-activated webhook %s") % obj)

    @admin.action(description=_("Purge queued messages from selected webhooks"))
    def purge_messages(self, request, queryset):
        actions_count = 0
        killmails_deleted = 0
        for webhook in queryset:
            killmails_deleted += webhook.clear_queue()
            actions_count += 1
        self.message_user(
            request,
            _(
                "Purged queued messages for %(actions_count)s webhooks, "
                "deleting a total of %(killmails_deleted)s messages."
            )
            % {"actions_count": actions_count, "killmails_deleted": killmails_deleted},
        )

    @admin.action(description=_("Send queued messages from selected webhooks"))
    def send_messages(self, request, queryset):
        items_count = 0
        for webhook in queryset:
            tasks.send_messages_for_webhook.delay(webhook.pk)  # type: ignore
            items_count += 1

        self.message_user(
            request, _("Started sending queued messages for %d webhooks.") % items_count
        )



"""Common logic used by all views."""

from typing import Optional

from structures.app_settings import (
    STRUCTURES_DEFAULT_PAGE_LENGTH,
    STRUCTURES_PAGING_ENABLED,
)
from structures.models import Owner


def add_common_context(context: Optional[dict] = None) -> dict:
    """Add common context and return it."""
    new_context = {
        "last_updated": Owner.objects.structures_last_updated(),
    }
    if context:
        new_context.update(context)
    return new_context


def add_common_data_export(data_export: dict) -> dict:
    """Add common data export entries to dictionary and return it."""
    add_on = {
        "data_tables_page_length": STRUCTURES_DEFAULT_PAGE_LENGTH,
        "data_tables_paging": int(STRUCTURES_PAGING_ENABLED),
    }
    data_export.update(add_on)
    return data_export



"""Status views."""

from django.http import HttpRequest, HttpResponse, HttpResponseServerError

from structures.models import Owner


def service_status(request: HttpRequest):
    """Public view to 3rd party monitoring.

    This is view allows running a 3rd party monitoring on the status
    of this services. Service will be reported as down if any of the
    configured structure or notifications syncs fails or is delayed
    """
    active_owners = Owner.objects.filter(
        is_included_in_service_status=True, is_active=True
    )
    for owner in active_owners:
        if not owner.are_all_syncs_ok:
            return HttpResponseServerError("service is down")

    return HttpResponse("service is up")



from unittest.mock import patch

from app_utils.esi_testing import EsiClientStub, EsiEndpoint
from app_utils.testing import NoSocketsTestCase

from structures.constants import EveCorporationId
from structures.models import PocoDetails, Structure
from structures.tests.testdata.factories import (
    EveEntityCorporationFactory,
    OwnerFactory,
    PocoFactory,
    UserMainDefaultOwnerFactory,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.models.owners"


@patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
@patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
@patch(MODULE_PATH + ".esi")
class TestUpdatePocosEsi(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user, structures_last_update_at=None)
        cls.corporation_id = cls.owner.corporation.corporation_id
        EveEntityCorporationFactory(
            id=EveCorporationId.DED, name="DED"
        )  # for notifications
        cls.endpoints = [
            EsiEndpoint(
                "Assets",
                "post_corporations_corporation_id_assets_locations",
                "corporation_id",
                needs_token=True,
                data={
                    str(cls.corporation_id): [
                        {
                            "item_id": 1200000000003,
                            "position": {"x": 1.2, "y": 2.3, "z": -3.4},
                        },
                        {
                            "item_id": 1200000000004,
                            "position": {"x": 5.2, "y": 6.3, "z": -7.4},
                        },
                        {
                            "item_id": 1200000000005,
                            "position": {"x": 1.2, "y": 6.3, "z": -7.4},
                        },
                        {
                            "item_id": 1200000000006,
                            "position": {"x": 41.2, "y": 26.3, "z": -47.4},
                        },
                    ]
                },
            ),
            EsiEndpoint(
                "Assets",
                "post_corporations_corporation_id_assets_names",
                "corporation_id",
                needs_token=True,
                data={
                    str(cls.corporation_id): [
                        {
                            "item_id": 1200000000003,
                            "name": "Customs Office (Amamake V)",
                        },
                        {
                            "item_id": 1200000000004,
                            "name": "Customs Office (1-PGSG VI)",
                        },
                        {
                            "item_id": 1200000000005,
                            "name": "Customs Office (1-PGSG VII)",
                        },
                        {
                            "item_id": 1200000000006,
                            "name": '<localized hint="Customs Office">Customs Office*</localized> (1-PGSG VIII)',
                        },
                    ]
                },
            ),
            EsiEndpoint(
                "Planetary_Interaction",
                "get_corporations_corporation_id_customs_offices",
                "corporation_id",
                needs_token=True,
                data={
                    str(cls.corporation_id): [
                        {
                            "alliance_tax_rate": 0.02,
                            "allow_access_with_standings": True,
                            "allow_alliance_access": True,
                            "bad_standing_tax_rate": 0.3,
                            "corporation_tax_rate": 0.02,
                            "excellent_standing_tax_rate": 0.02,
                            "good_standing_tax_rate": 0.02,
                            "neutral_standing_tax_rate": 0.02,
                            "office_id": 1200000000003,
                            "reinforce_exit_end": 21,
                            "reinforce_exit_start": 19,
                            "standing_level": "terrible",
                            "system_id": 30002537,
                            "terrible_standing_tax_rate": 0.5,
                        },
                        {
                            "alliance_tax_rate": 0.02,
                            "allow_access_with_standings": True,
                            "allow_alliance_access": True,
                            "bad_standing_tax_rate": 0.02,
                            "corporation_tax_rate": 0.02,
                            "excellent_standing_tax_rate": 0.02,
                            "good_standing_tax_rate": 0.02,
                            "neutral_standing_tax_rate": 0.02,
                            "office_id": 1200000000004,
                            "reinforce_exit_end": 21,
                            "reinforce_exit_start": 19,
                            "standing_level": "terrible",
                            "system_id": 30000474,
                            "terrible_standing_tax_rate": 0.02,
                        },
                        {
                            "alliance_tax_rate": 0.02,
                            "allow_access_with_standings": True,
                            "allow_alliance_access": True,
                            "bad_standing_tax_rate": 0.02,
                            "corporation_tax_rate": 0.02,
                            "excellent_standing_tax_rate": 0.02,
                            "good_standing_tax_rate": 0.02,
                            "neutral_standing_tax_rate": 0.02,
                            "office_id": 1200000000005,
                            "reinforce_exit_end": 21,
                            "reinforce_exit_start": 19,
                            "standing_level": "terrible",
                            "system_id": 30000474,
                            "terrible_standing_tax_rate": 0.02,
                        },
                        {
                            "alliance_tax_rate": 0.02,
                            "allow_access_with_standings": True,
                            "allow_alliance_access": True,
                            "bad_standing_tax_rate": 0.02,
                            "corporation_tax_rate": 0.02,
                            "excellent_standing_tax_rate": 0.02,
                            "good_standing_tax_rate": 0.02,
                            "neutral_standing_tax_rate": 0.02,
                            "office_id": 1200000000006,
                            "reinforce_exit_end": 21,
                            "reinforce_exit_start": 19,
                            "standing_level": "terrible",
                            "system_id": 30000474,
                            "terrible_standing_tax_rate": 0.02,
                        },
                        {
                            "alliance_tax_rate": 0.02,
                            "allow_access_with_standings": True,
                            "allow_alliance_access": True,
                            "bad_standing_tax_rate": 0.02,
                            "corporation_tax_rate": 0.02,
                            "excellent_standing_tax_rate": 0.02,
                            "good_standing_tax_rate": 0.02,
                            "neutral_standing_tax_rate": 0.02,
                            "office_id": 1200000000099,
                            "reinforce_exit_end": 21,
                            "reinforce_exit_start": 19,
                            "standing_level": "terrible",
                            "system_id": 30000474,
                            "terrible_standing_tax_rate": 0.02,
                        },
                    ]
                },
            ),
            EsiEndpoint(
                "Corporation",
                "get_corporations_corporation_id_structures",
                "corporation_id",
                needs_token=True,
                data={str(cls.corporation_id): [], "2005": []},
            ),  # TODO: Remove once possible
        ]
        cls.esi_client_stub = EsiClientStub.create_from_endpoints(cls.endpoints)

    def test_can_sync_pocos(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)

        # must contain all expected structures
        expected = {
            1200000000003,
            1200000000004,
            1200000000005,
            1200000000006,
            1200000000099,
        }
        self.assertSetEqual(owner.structures.ids(), expected)
        self.assertSetEqual(
            set(PocoDetails.objects.values_list("structure_id", flat=True)),
            {1200000000003, 1200000000004, 1200000000005, 1200000000006, 1200000000099},
        )

        # verify attributes for POCO
        structure = Structure.objects.get(id=1200000000003)
        self.assertEqual(structure.name, "Planet (Barren)")
        self.assertEqual(structure.eve_solar_system_id, 30002537)
        self.assertEqual(
            int(structure.owner.corporation.corporation_id), self.corporation_id
        )
        self.assertEqual(structure.eve_type_id, 2233)
        self.assertEqual(structure.reinforce_hour, 20)
        self.assertEqual(structure.state, Structure.State.UNKNOWN)
        self.assertEqual(structure.eve_planet_id, 40161472)

        # verify attributes for POCO details
        details = structure.poco_details
        self.assertEqual(details.alliance_tax_rate, 0.02)
        self.assertTrue(details.allow_access_with_standings)
        self.assertTrue(details.allow_alliance_access)
        self.assertEqual(details.bad_standing_tax_rate, 0.3)
        self.assertEqual(details.corporation_tax_rate, 0.02)
        self.assertEqual(details.excellent_standing_tax_rate, 0.02)
        self.assertEqual(details.good_standing_tax_rate, 0.02)
        self.assertEqual(details.neutral_standing_tax_rate, 0.02)
        self.assertEqual(details.reinforce_exit_end, 21)
        self.assertEqual(details.reinforce_exit_start, 19)
        self.assertEqual(details.standing_level, PocoDetails.StandingLevel.TERRIBLE)
        self.assertEqual(details.terrible_standing_tax_rate, 0.5)

        # empty name for POCO with no asset data
        structure = Structure.objects.get(id=1200000000099)
        self.assertEqual(structure.name, "")

    def test_should_not_break_on_http_error_when_fetching_custom_offices(
        self, mock_esi
    ):
        # given
        new_endpoint = EsiEndpoint(
            "Planetary_Interaction",
            "get_corporations_corporation_id_customs_offices",
            http_error_code=500,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        expected = set()
        self.assertSetEqual(owner.structures.ids(), expected)

    def test_should_not_break_on_http_error_when_fetching_custom_office_names(
        self, mock_esi
    ):
        # given
        new_endpoint = EsiEndpoint(
            "Assets",
            "post_corporations_corporation_id_assets_names",
            http_error_code=404,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()
        # then
        expected = {
            1200000000003,
            1200000000004,
            1200000000005,
            1200000000006,
            1200000000099,
        }
        self.assertSetEqual(owner.structures.ids(), expected)

    def test_should_remove_old_pocos(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        PocoFactory(owner=owner, id=1200000000010, name="delete-me")
        # when
        owner.update_structures_esi()
        # then
        expected = {
            1200000000003,
            1200000000004,
            1200000000005,
            1200000000006,
            1200000000099,
        }
        self.assertSetEqual(owner.structures.ids(), expected)

    def test_should_not_delete_existing_pocos_when_update_failed(self, mock_esi):
        # given
        new_endpoint = EsiEndpoint(
            "Planetary_Interaction",
            "get_corporations_corporation_id_customs_offices",
            http_error_code=500,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        PocoFactory(owner=owner, id=1200000000003)
        PocoFactory(owner=owner, id=1200000000004)
        # when
        owner.update_structures_esi()
        # then
        self.assertFalse(owner.is_structure_sync_fresh)
        expected = {1200000000003, 1200000000004}
        self.assertSetEqual(owner.structures.ids(), expected)

    def test_should_have_empty_name_if_not_match_with_planets(self, mock_esi):
        # given
        owner = OwnerFactory(structures_last_update_at=None)
        corporation_id = owner.corporation.corporation_id
        endpoints = [
            EsiEndpoint(
                "Assets",
                "post_corporations_corporation_id_assets_locations",
                "corporation_id",
                needs_token=True,
                data={
                    f"{corporation_id}": [
                        {
                            "item_id": 1200000000099,
                            "position": {"x": 1.2, "y": 2.3, "z": -3.4},
                        }
                    ]
                },
            ),
            EsiEndpoint(
                "Assets",
                "post_corporations_corporation_id_assets_names",
                "corporation_id",
                needs_token=True,
                data={
                    f"{corporation_id}": [
                        {
                            "item_id": 1200000000099,
                            "name": "Invalid name",
                        }
                    ]
                },
            ),
            EsiEndpoint(
                "Planetary_Interaction",
                "get_corporations_corporation_id_customs_offices",
                "corporation_id",
                needs_token=True,
                data={
                    f"{corporation_id}": [
                        {
                            "alliance_tax_rate": 0.02,
                            "allow_access_with_standings": True,
                            "allow_alliance_access": True,
                            "bad_standing_tax_rate": 0.3,
                            "corporation_tax_rate": 0.02,
                            "excellent_standing_tax_rate": 0.02,
                            "good_standing_tax_rate": 0.02,
                            "neutral_standing_tax_rate": 0.02,
                            "office_id": 1200000000099,
                            "reinforce_exit_end": 21,
                            "reinforce_exit_start": 19,
                            "standing_level": "terrible",
                            "system_id": 30002537,
                            "terrible_standing_tax_rate": 0.5,
                        }
                    ]
                },
            ),
            EsiEndpoint(
                "Corporation",
                "get_corporations_corporation_id_structures",
                "corporation_id",
                needs_token=True,
                data={f"{corporation_id}": []},
            ),  # TODO: Remove once possible
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # when
        owner.update_structures_esi()
        # then
        self.assertTrue(owner.is_structure_sync_fresh)
        structure = Structure.objects.get(id=1200000000099)
        self.assertEqual(structure.name, "")

    def test_define_poco_name_from_planet_type_if_found(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()
        # then
        structure = Structure.objects.get(id=1200000000003)
        self.assertEqual(structure.eve_planet_id, 40161472)
        self.assertEqual(structure.name, "Planet (Barren)")



from django.test import RequestFactory

from app_utils.testing import NoSocketsTestCase

from structures.views import public

from ..testdata.factories import OwnerFactory, PocoFactory, UserMainBasicFactory
from ..testdata.load_eveuniverse import load_eveuniverse
from .utils import json_response_to_dict


class TestPocoListDataView(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        cls.user = UserMainBasicFactory()
        cls.main = cls.user.profile.main_character
        owner = OwnerFactory(are_pocos_public=True)
        cls.poco_public = PocoFactory(
            owner=owner,
            eve_planet_name="Amamake V",
            poco_details__allow_access_with_standings=True,
            poco_details__neutral_standing_tax_rate=0.01,
        )
        cls.poco_non_public = PocoFactory()

    def test_should_return_public_pocos_only(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = public.public_poco_list_data(request, self.main.character_id)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        structure_ids = set(data.keys())
        self.assertSetEqual(structure_ids, {self.poco_public.id})



from random import randrange


def get_random_subset(lst: list, max_members: int) -> list:
    lst2 = lst.copy()
    subset = list()
    max_members = min(max_members, len(lst))

    for x in range(randrange(max_members) + 1):
        m = lst2.pop(randrange(len(lst2)))
        subset.append(m)

    return subset


my_list = ["one", "two", "three", "four"]

subset = get_random_subset(my_list, 4)
# print(subset)



import datetime as dt
from unittest.mock import patch

from django.test import TestCase
from django.urls import reverse
from django.utils.timezone import now

from app_utils.testing import response_text

from structures.tests.testdata.factories import OwnerFactory
from structures.views import status

OWNERS_PATH = "structures.models.owners"


@patch(OWNERS_PATH + ".STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES", 30)
@patch(OWNERS_PATH + ".STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES", 30)
class TestServiceStatus(TestCase):
    def test_should_report_ok(self):
        # given
        OwnerFactory(
            structures_last_update_at=now(),
            notifications_last_update_at=now(),
            forwarding_last_update_at=now(),
            assets_last_update_at=now(),
        )
        request = self.client.get(reverse("structures:service_status"))
        # when
        response = status.service_status(request)
        # then
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response_text(response), "service is up")

    def test_should_report_ok_when_downed_owner_is_inactive(self):
        # given
        OwnerFactory(
            structures_last_update_at=now(),
            notifications_last_update_at=now(),
            forwarding_last_update_at=now(),
            assets_last_update_at=now(),
        )
        OwnerFactory(
            structures_last_update_at=now() - dt.timedelta(minutes=31),
            notifications_last_update_at=now(),
            forwarding_last_update_at=now(),
            assets_last_update_at=now(),
            is_active=False,
        )
        request = self.client.get(reverse("structures:service_status"))
        # when
        response = status.service_status(request)
        # then
        self.assertEqual(response.status_code, 200)

    def test_should_report_fail_when_issue_with_structures(self):
        # given
        OwnerFactory(
            structures_last_update_at=now() - dt.timedelta(minutes=31),
            notifications_last_update_at=now(),
            forwarding_last_update_at=now(),
            assets_last_update_at=now(),
        )
        request = self.client.get(reverse("structures:service_status"))
        # when
        response = status.service_status(request)
        # then
        self.assertEqual(response.status_code, 500)
        self.assertEqual(response_text(response), "service is down")

    def test_should_report_fail_when_issue_with_notifications(self):
        # given
        OwnerFactory(
            structures_last_update_at=now(),
            notifications_last_update_at=now() - dt.timedelta(minutes=31),
            forwarding_last_update_at=now(),
            assets_last_update_at=now(),
        )
        request = self.client.get(reverse("structures:service_status"))
        # when
        response = status.service_status(request)
        # then
        self.assertEqual(response.status_code, 500)
        self.assertEqual(response_text(response), "service is down")

    def test_should_report_fail_when_issue_with_forwarding(self):
        # given
        OwnerFactory(
            structures_last_update_at=now(),
            notifications_last_update_at=now(),
            forwarding_last_update_at=now() - dt.timedelta(minutes=31),
            assets_last_update_at=now(),
        )
        request = self.client.get(reverse("structures:service_status"))
        # when
        response = status.service_status(request)
        # then
        self.assertEqual(response.status_code, 500)
        self.assertEqual(response_text(response), "service is down")

    def test_should_report_fail_when_issue_with_assets(self):
        # given
        OwnerFactory(
            structures_last_update_at=now(),
            notifications_last_update_at=now(),
            forwarding_last_update_at=now(),
            assets_last_update_at=now() - dt.timedelta(minutes=31),
        )
        request = self.client.get(reverse("structures:service_status"))
        # when
        response = status.service_status(request)
        # then
        self.assertEqual(response.status_code, 500)
        self.assertEqual(response_text(response), "service is down")



"""Tasks for Structures."""

from django.urls import path
from django.views.decorators.cache import never_cache

from .views import public, statistics, status, structures

app_name = "structures"

urlpatterns = [
    path("", structures.index, name="index"),
    path("list", structures.structure_list, name="structure_list"),
    path(
        "structure_list_data/<str:selection>",
        structures.structure_list_data,
        name="structure_list_data",
    ),
    path(
        "add_structure_owner",
        structures.add_structure_owner,
        name="add_structure_owner",
    ),
    path("service_status", never_cache(status.service_status), name="service_status"),
    path(
        "<int:structure_id>/structure_details",
        structures.structure_details,
        name="structure_details",
    ),
    path(
        "<int:structure_id>/poco_details",
        structures.poco_details,
        name="poco_details",
    ),
    path(
        "<int:structure_id>/starbase_detail",
        structures.starbase_detail,
        name="starbase_detail",
    ),
    # public
    path("public", public.public, name="public"),
    path(
        "public_poco_list_data/<int:character_id>",
        public.public_poco_list_data,
        name="public_poco_list_data",
    ),
    # statistics
    path("statistics", statistics.statistics, name="statistics"),
    path(
        "summary_data", statistics.structure_summary_data, name="structure_summary_data"
    ),
]



import datetime as dt
from unittest.mock import patch

from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase

from structures.core.notification_types import NotificationType
from structures.models import GeneratedNotification, Notification, Owner

from .testdata.factories import (
    GeneratedNotificationFactory,
    NotificationFactory,
    OwnerCharacterFactory,
    OwnerFactory,
)
from .testdata.load_eveuniverse import load_eveuniverse

MANAGERS_PATH = "structures.managers"


@patch(
    "structures.models.notifications.NotificationBase.add_or_remove_timer",
    spec=True,
)
class TestNotificationBaseAddOrRemoveTimers(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_create_new_timers_from_notifications(
        self, mock_add_or_remove_timer_from_notification
    ):
        # given
        owner = OwnerFactory()
        NotificationFactory(
            owner=owner, notif_type=NotificationType.STRUCTURE_LOST_SHIELD
        )
        NotificationFactory(
            owner=owner, notif_type=NotificationType.WAR_CORPORATION_BECAME_ELIGIBLE
        )
        # when
        Notification.objects.add_or_remove_timers()
        # then
        self.assertEqual(mock_add_or_remove_timer_from_notification.call_count, 1)

    def test_should_create_new_timers_from_generated_notifications(
        self, mock_add_or_remove_timer_from_notification
    ):
        # given
        GeneratedNotificationFactory()
        # when
        GeneratedNotification.objects.add_or_remove_timers()
        # then
        self.assertEqual(mock_add_or_remove_timer_from_notification.call_count, 1)


class TestOwnerManager(NoSocketsTestCase):
    def test_should_annotate_characters_count(self):
        # given
        owner = OwnerFactory()  # 1st character automatically created
        OwnerCharacterFactory(owner=owner, is_enabled=False)  # 2nd character added
        # when
        result = Owner.objects.annotate_characters_count()
        # then
        obj = result.get(pk=owner.pk)
        self.assertEqual(obj.characters_enabled_count, 1)
        self.assertEqual(obj.characters_disabled_count, 1)

    def test_should_return_when_structures_where_last_updated_for_several_owners(self):
        # given
        owner_1_structures_last_update_at = now() - dt.timedelta(hours=3)
        OwnerFactory(structures_last_update_at=owner_1_structures_last_update_at)
        owner_2_structures_last_update_at = now() - dt.timedelta(hours=1)
        OwnerFactory(structures_last_update_at=owner_2_structures_last_update_at)
        # when
        result = Owner.objects.structures_last_updated()
        # then
        self.assertEqual(result, owner_2_structures_last_update_at)

    def test_should_return_none_when_structures_where_last_updated_and_no_owners(
        self,
    ):
        # when
        result = Owner.objects.structures_last_updated()
        # then
        self.assertIsNone(result)



from app_utils.testing import json_response_to_python


def json_response_to_dict(response, key="id") -> dict:
    """Convert JSON response into dict by given key."""
    return {obj[key]: obj for obj in json_response_to_python(response)["data"]}



from unittest.mock import patch

from django.contrib.auth.models import User
from django.test import TestCase, override_settings

from app_utils.testdata_factories import UserFactory
from app_utils.testing import NoSocketsTestCase, generate_invalid_pk

from structures import tasks
from structures.core.notification_types import NotificationType
from structures.models import Owner, Webhook

from .testdata.factories import (
    FuelAlertConfigFactory,
    JumpFuelAlertConfigFactory,
    NotificationFactory,
    OwnerFactory,
    UserMainDefaultOwnerFactory,
    WebhookFactory,
)

MODULE_PATH = "structures.tasks"
MODULE_PATH_MODELS_OWNERS = "structures.models.owners"


@patch(MODULE_PATH + ".Webhook.send_queued_messages", spec=True)
class TestSendMessagesForWebhook(TestCase):
    @classmethod
    def setUpClass(cls):  # Can not be setUpTestData due to conflict with redis client
        super().setUpClass()
        cls.webhook = WebhookFactory()

    def test_normal(self, mock_send_queued_messages):
        tasks.send_messages_for_webhook(self.webhook.pk)
        self.assertEqual(mock_send_queued_messages.call_count, 1)

    def test_invalid_pk(self, mock_send_queued_messages):
        tasks.send_messages_for_webhook(generate_invalid_pk(Webhook))
        self.assertEqual(mock_send_queued_messages.call_count, 0)

    def test_disabled_webhook(self, mock_send_queued_messages):
        self.webhook.is_active = False
        self.webhook.save()

        tasks.send_messages_for_webhook(self.webhook.pk)
        self.assertEqual(mock_send_queued_messages.call_count, 0)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".Owner.update_structures_esi", spec=True)
class TestUpdateStructuresEsi(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user, is_alliance_main=True)

    def test_call_structure_update_with_owner_only(self, mock_update_structures_esi):
        """TODO: Investigate how to call the top level method that contains the chains()"""
        tasks.update_structures_esi_for_owner(self.owner.pk)
        self.assertTrue(mock_update_structures_esi.called)

    def test_call_structure_update_with_owner_and_user(
        self, mock_update_structures_esi
    ):
        """TODO: Investigate how to call the top level method that contains the chains()"""
        tasks.update_structures_esi_for_owner(self.owner.pk, self.user.pk)
        first, _ = mock_update_structures_esi.call_args
        self.assertEqual(first[0], self.user)

    def test_call_structure_update_with_owner_and_ignores_invalid_user(
        self, mock_update_structures_esi
    ):
        """TODO: Investigate how to call the top level method that contains the chains()"""
        tasks.update_structures_esi_for_owner(self.owner.pk, generate_invalid_pk(User))
        first, _ = mock_update_structures_esi.call_args
        self.assertIsNone(first[0])

    def test_raises_exception_if_owner_is_unknown(self, mock_update_structures_esi):
        with self.assertRaises(Owner.DoesNotExist):
            """TODO: Investigate how to call the top level method that contains the chains()"""
            tasks.update_structures_esi_for_owner(owner_pk=generate_invalid_pk(Owner))


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".update_structures_for_owner", spec=True)
class TestUpdateStructuresForOwner(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user, is_alliance_main=True)

    def test_can_update_structures_for_all_owners(
        self, mock_update_structures_for_owner
    ):
        # given
        owner_2 = OwnerFactory()
        # when
        tasks.update_structures()
        # then
        call_args_list = mock_update_structures_for_owner.delay.call_args_list
        owner_pks = {args[0][0] for args in call_args_list}
        self.assertSetEqual(owner_pks, {self.owner.pk, owner_2.pk})

    def test_does_not_update_structures_for_non_active_owners(
        self, mock_update_structures_for_owner
    ):
        # given
        OwnerFactory(is_active=False)
        # when
        tasks.update_structures()
        # then
        call_args_list = mock_update_structures_for_owner.delay.call_args_list
        owner_pks = {args[0][0] for args in call_args_list}
        self.assertSetEqual(owner_pks, {self.owner.pk})


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestUpdateOwnerAsset(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user, is_alliance_main=True)

    @patch(MODULE_PATH + ".Owner.update_asset_esi")
    def test_call_structure_asset_update_with_owner_and_user(
        self, mock_update_asset_esi
    ):
        """TODO: Investigate how to call the top level method that contains the chains()"""
        tasks.update_structures_assets_for_owner(self.owner.pk, self.user.pk)
        first, _ = mock_update_asset_esi.call_args
        self.assertEqual(first[0], self.user)

    @patch(MODULE_PATH + ".Owner.update_asset_esi")
    def test_call_structure_asset_update_with_owner_and_ignores_invalid_user(
        self, mock_update_asset_esi
    ):
        """TODO: Investigate how to call the top level method that contains the chains()"""
        tasks.update_structures_assets_for_owner(
            self.owner.pk, generate_invalid_pk(User)
        )
        first, _ = mock_update_asset_esi.call_args
        self.assertIsNone(first[0])

    @override_settings(
        CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True
    )
    def test_raises_exception_if_owner_is_unknown(self):
        with self.assertRaises(Owner.DoesNotExist):
            """TODO: Investigate how to call the top level method that contains the chains()"""
            tasks.update_structures_assets_for_owner(
                owner_pk=generate_invalid_pk(Owner)
            )


@patch(MODULE_PATH_MODELS_OWNERS + ".Owner.update_is_up", lambda *args, **kwargs: None)
@patch(MODULE_PATH + ".send_structure_fuel_notifications_for_config", spec=True)
@patch(MODULE_PATH + ".process_notifications_for_owner", spec=True)
class TestFetchAllNotifications(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user, is_alliance_main=True)

    def test_fetch_all_notifications(
        self, mock_fetch_notifications_owner, mock_send_fuel_notifications_for_config
    ):
        # given
        owner_2 = OwnerFactory()
        OwnerFactory(is_active=False)
        # when
        tasks.fetch_all_notifications()
        # then
        call_args_list = mock_fetch_notifications_owner.apply_async.call_args_list
        owner_pks = {obj[1]["kwargs"]["owner_pk"] for obj in call_args_list}
        self.assertSetEqual(owner_pks, {self.owner.pk, owner_2.pk})

    def test_send_new_fuel_notifications(
        self, mock_fetch_notifications_owner, mock_send_fuel_notifications_for_config
    ):
        # given
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        # when
        tasks.fetch_all_notifications()
        # then
        self.assertEqual(mock_send_fuel_notifications_for_config.delay.call_count, 1)
        args, _ = mock_send_fuel_notifications_for_config.delay.call_args
        self.assertEqual(args[0], config.pk)


# TODO: Fix tests. Does not work with tox.
# @override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
# @patch(MODULE_PATH + ".fetch_esi_status", lambda: EsiStatus(True, 100, 60))
# class TestProcessNotificationsForOwner(TestCase):
#     @classmethod
#     def setUpClass(cls) -> None:
#         super().setUpClass()
#         create_structures()
#         cls.user, cls.owner = set_owner_character(character_id=1001)

#     def test_should_raise_exception_if_owner_is_unknown(self):
#         with self.assertRaises(Owner.DoesNotExist):
#             tasks.process_notifications_for_owner.delay(
#                 owner_pk=generate_invalid_pk(Owner)
#             )

#     @patch(MODULE_PATH + ".send_messages_for_webhook")
#     @patch(MODULE_PATH + ".Owner.fetch_notifications_esi")
#     def test_should_send_notifications_for_owner(
#         self, mock_fetch_notifications_esi, mock_send_messages_for_webhook
#     ):
#         # given
#         load_notification_entities(self.owner)
#         Notification.objects.exclude(notification_id=1000000509).delete()
#         self.owner.webhooks.first().clear_queue()
#         # when
#         tasks.process_notifications_for_owner.delay(owner_pk=self.owner.pk)
#         # then
#         self.assertTrue(mock_fetch_notifications_esi.called)
#         self.assertEqual(mock_send_messages_for_webhook.apply_async.call_count, 1)
#         for notif in self.owner.notifications.filter(
#             notif_type__in=[NotificationType.structure_related()]
#         ):
#             structure_ids = notif.structures.values_list("id", flat=True)
#             self.assertTrue(
#                 1000000000001 in set(structure_ids)
#                 or 1000000000002 in set(structure_ids)
#             )

#     @patch(MODULE_PATH + ".send_messages_for_webhook")
#     @patch(MODULE_PATH + ".Owner.fetch_notifications_esi")
#     def test_dont_sent_if_queue_is_empty(
#         self, mock_fetch_notifications_esi, mock_send_messages_for_webhook
#     ):
#         # given
#         self.owner.webhooks.first().clear_queue()
#         # when
#         tasks.process_notifications_for_owner.delay(owner_pk=self.owner.pk)
#         # then
#         self.assertTrue(mock_fetch_notifications_esi.called)
#         self.assertEqual(mock_send_messages_for_webhook.apply_async.call_count, 0)


@patch("structures.webhooks.core.sleep", lambda _: None)
@patch(MODULE_PATH + ".notify", spec=True)
@patch("structures.models.notifications.Webhook.send_test_message", spec=True)
class TestSendTestNotification(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user, is_alliance_main=True)

    def test_send_test_notification(self, mock_send_test_message, mock_notify):
        # given
        mock_send_test_message.return_value = ("", True)
        my_webhook = self.owner.webhooks.first()

        # when
        tasks.send_test_notifications_to_webhook(my_webhook.pk, self.user.pk)

        # then
        self.assertEqual(mock_send_test_message.call_count, 1)
        self.assertTrue(mock_notify.called)
        args = mock_notify.call_args[1]
        self.assertEqual(args["level"], "success")

    def test_send_test_notification_error(self, mock_send_test_message, mock_notify):
        # given
        mock_send_test_message.return_value = ("Error", False)
        my_webhook = self.owner.webhooks.first()

        # when
        tasks.send_test_notifications_to_webhook(my_webhook.pk, self.user.pk)

        # then
        self.assertEqual(mock_send_test_message.call_count, 1)
        self.assertTrue(mock_notify.called)
        args = mock_notify.call_args[1]
        self.assertEqual(args["level"], "danger")


@patch("structures.models.notifications.Notification.update_related_structures")
class TestUpdateNotificationsStructureRelations(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.owner = OwnerFactory(is_alliance_main=True)

    def test_should_run_updates(self, mock_update_related_structures):
        # given
        mock_update_related_structures.return_value = True
        NotificationFactory(
            owner=self.owner, notif_type=NotificationType.STRUCTURE_UNDER_ATTACK
        )
        NotificationFactory(
            owner=self.owner, notif_type=NotificationType.CORP_APP_NEW_MSG
        )
        # when
        result = tasks.update_notifications_structure_relations(self.owner.pk)
        # then
        self.assertEqual(result, 1)

    def test_should_run_no_updates(self, mock_update_related_structures):
        # given
        mock_update_related_structures.return_value = True
        NotificationFactory(
            owner=self.owner, notif_type=NotificationType.CORP_APP_NEW_MSG
        )
        # when
        result = tasks.update_notifications_structure_relations(self.owner.pk)
        # then
        self.assertEqual(result, 0)


class TestOtherTasks(NoSocketsTestCase):
    @patch(
        MODULE_PATH + ".EveSovereigntyMap.objects.update_or_create_all_from_esi",
        spec=True,
    )
    def test_should_call_update_sov_map_from_esi(self, mock_update_from_esi):
        # when
        tasks.update_sov_map()
        # then
        self.assertTrue(mock_update_from_esi.called)

    @patch(MODULE_PATH + ".Owner.fetch_notifications_esi", spec=True)
    def test_should_fetch_notifications_for_owner(self, mock_fetch_notifications_esi):
        # given
        owner = OwnerFactory()
        # when
        tasks.fetch_notification_for_owner(owner.pk)
        # then
        self.assertTrue(mock_fetch_notifications_esi.called)

    @patch(MODULE_PATH + ".send_queued_messages_for_webhooks", spec=True)
    @patch(MODULE_PATH + ".Owner.send_new_notifications", spec=True)
    def test_should_send_notifications_for_owner(
        self, mock_send_new_notifications, mock_send_queued_messages_for_webhooks
    ):
        # given
        owner = OwnerFactory()
        # when
        tasks.send_new_notifications_for_owner(owner.pk)
        # then
        self.assertTrue(mock_send_new_notifications.called)
        self.assertTrue(mock_send_queued_messages_for_webhooks.called)

    @patch(MODULE_PATH + ".send_queued_messages_for_webhooks", spec=True)
    @patch(MODULE_PATH + ".FuelAlertConfig.send_new_notifications", spec=True)
    def test_should_send_fuel_notifications(
        self, mock_send_new_notifications, mock_send_queued_messages_for_webhooks
    ):
        # given
        config = FuelAlertConfigFactory()
        # when
        tasks.send_structure_fuel_notifications_for_config(config.pk)
        # then
        self.assertTrue(mock_send_new_notifications.called)
        self.assertTrue(mock_send_queued_messages_for_webhooks.called)

    @patch(MODULE_PATH + ".send_queued_messages_for_webhooks", spec=True)
    @patch(MODULE_PATH + ".JumpFuelAlertConfig.send_new_notifications", spec=True)
    def test_should_send_jump_fuel_notifications(
        self, mock_send_new_notifications, mock_send_queued_messages_for_webhooks
    ):
        # given
        config = JumpFuelAlertConfigFactory()
        # when
        tasks.send_jump_fuel_notifications_for_config(config.pk)
        # then
        self.assertTrue(mock_send_new_notifications.called)
        self.assertTrue(mock_send_queued_messages_for_webhooks.called)

    @patch(MODULE_PATH + ".send_messages_for_webhook", spec=True)
    @patch(MODULE_PATH + ".Webhook.queue_size", spec=True)
    def test_should_send_queued_messages_to_webhooks_1(
        self, mock_queue_size, mock_send_messages_for_webhook
    ):
        # given
        mock_queue_size.return_value = 1
        webhook_1 = WebhookFactory()
        webhook_2 = WebhookFactory()
        # when
        tasks.send_queued_messages_for_webhooks([webhook_1, webhook_2])
        # then
        called_webhook_pks = {
            obj[1]["kwargs"]["webhook_pk"]
            for obj in mock_send_messages_for_webhook.apply_async.call_args_list
        }
        expected = {webhook_1.pk, webhook_2.pk}
        self.assertSetEqual(called_webhook_pks, expected)

    @patch(MODULE_PATH + ".send_messages_for_webhook", spec=True)
    @patch(MODULE_PATH + ".Webhook.queue_size", spec=True)
    def test_should_send_queued_messages_to_webhooks_2(
        self, mock_queue_size, mock_send_messages_for_webhook
    ):
        # given
        mock_queue_size.return_value = 0
        webhook_1 = WebhookFactory()
        webhook_2 = WebhookFactory()
        # when
        tasks.send_queued_messages_for_webhooks([webhook_1, webhook_2])
        # then
        called_webhook_pks = {
            obj[1]["kwargs"]["webhook_pk"]
            for obj in mock_send_messages_for_webhook.apply_async.call_args_list
        }
        expected = set()
        self.assertSetEqual(called_webhook_pks, expected)


class TestGetUser(NoSocketsTestCase):
    def test_should_return_user(self):
        # given
        user = UserFactory()
        # when
        result = tasks._get_user(user.pk)
        # then
        self.assertEqual(result, user)

    def test_should_return_none_when_not_found(self):
        # when
        result = tasks._get_user(generate_invalid_pk(User))
        # then
        self.assertIsNone(result)

    def test_should_return_none_when_called_with_none(self):
        # when
        result = tasks._get_user(None)
        # then
        self.assertIsNone(result)



from app_utils.testing import NoSocketsTestCase

from structures.constants import EveTypeId
from structures.core import sovereignty


class TestSovereignty(NoSocketsTestCase):
    def test_should_return_type_id_or_none(self):
        my_map = [(1, EveTypeId.TCU), (2, EveTypeId.IHUB), (3, None), (0, None)]
        for input, expected in my_map:
            with self.subTest(input=input):
                self.assertEqual(sovereignty.event_type_to_type_id(input), expected)

    def test_should_return_structure_type_names(self):
        my_map = [(1, "TCU"), (2, "I-HUB"), (3, "Other"), (0, "Other")]
        for input, expected in my_map:
            with self.subTest(input=input):
                self.assertEqual(
                    sovereignty.event_type_to_structure_type_name(input), expected
                )



from django.test import TestCase
from eveuniverse.models import EvePlanet, EveSolarSystem, EveType
from eveuniverse.tools.testdata import ModelSpec, create_testdata

from structures.constants import EveCategoryId, EveGroupId, EveTypeId

from .helpers import test_data_filename


class CreateEveUniverseTestData(TestCase):
    def test_create_testdata(self):
        testdata_spec = [
            ModelSpec(
                "EveCategory",
                ids=[EveCategoryId.STRUCTURE],
                include_children=True,
                enabled_sections=[EveType.Section.DOGMAS],
            ),
            ModelSpec(
                "EveCategory",
                ids=[EveCategoryId.ORBITAL],
                include_children=True,
            ),
            ModelSpec(
                "EveGroup",
                ids=[EveGroupId.CONTROL_TOWER],
                include_children=True,
                enabled_sections=[EveType.Section.DOGMAS],
            ),
            ModelSpec(
                "EveType",
                ids=[EveTypeId.TCU, EveTypeId.IHUB],
                include_children=True,
                enabled_sections=[EveType.Section.DOGMAS],
            ),
            ModelSpec(
                "EveGroup",
                ids=[
                    EveGroupId.PLANET,
                    EveGroupId.FUEL_BLOCK,
                    EveGroupId.ICE_PRODUCT,
                    EveGroupId.QUANTUM_CORES,
                    EveGroupId.STRUCTURE_CITADEL_SERVICE_MODULE,
                    EveGroupId.UNCOMMON_MOON_ASTEROIDS,
                ],
                include_children=True,
            ),
            ModelSpec("EveType", ids=[], include_children=False),
            ModelSpec(
                "EveSolarSystem",
                ids=[30002506, 31000005, 30002537, 30000474, 30000476],
                include_children=True,
                enabled_sections=[
                    EveSolarSystem.Section.PLANETS,
                    EvePlanet.Section.MOONS,
                ],
            ),
            ModelSpec(
                "EveMoon",
                ids=[40161465, 40161466, 40161471, 40029527],
                include_children=False,
            ),
        ]
        create_testdata(testdata_spec, test_data_filename())



import datetime as dt
from typing import Generic, List, Optional, TypeVar

import factory
import factory.fuzzy
import pytz
import yaml

from django.utils.text import slugify
from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveMoon, EvePlanet, EveSolarSystem, EveType

from allianceauth.eveonline.models import EveCharacter
from app_utils.testdata_factories import (
    EveAllianceInfoFactory,
    EveCharacterFactory,
    EveCorporationInfoFactory,
    UserFactory,
    UserMainFactory,
)

from structures.constants import EveGroupId, EveTypeId
from structures.core.notification_types import NotificationType
from structures.models import (
    EveSovereigntyMap,
    FuelAlert,
    FuelAlertConfig,
    GeneratedNotification,
    JumpFuelAlertConfig,
    Notification,
    Owner,
    OwnerCharacter,
    PocoDetails,
    StarbaseDetail,
    StarbaseDetailFuel,
    Structure,
    StructureItem,
    StructureService,
    StructureTag,
    Webhook,
)

# from .helpers import datetime_to_ldap  # TODO: Use for notifications

T = TypeVar("T")


class BaseMetaFactory(Generic[T], factory.base.FactoryMetaClass):
    def __call__(cls, *args, **kwargs) -> T:
        return super().__call__(*args, **kwargs)


# eve universe (within structures)


def datetime_to_esi(my_dt: dt.datetime) -> str:
    """Convert datetime to ESI datetime string."""
    utc_dt = my_dt.astimezone(pytz.utc)
    return utc_dt.strftime("%Y-%m-%dT%H:%M:%SZ")


class EveEntityFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[EveEntity]
):
    class Meta:
        model = EveEntity
        django_get_or_create = ("id",)

    category = EveEntity.CATEGORY_CHARACTER

    @factory.lazy_attribute
    def id(self):
        if self.category == EveEntity.CATEGORY_CHARACTER:
            obj = EveCharacterFactory()
            return obj.character_id
        if self.category == EveEntity.CATEGORY_CORPORATION:
            obj = EveCorporationInfoFactory()
            return obj.corporation_id
        if self.category == EveEntity.CATEGORY_ALLIANCE:
            obj = EveAllianceInfoFactory()
            return obj.alliance_id
        raise NotImplementedError(f"Unknown category: {self.category}")


class EveEntityCharacterFactory(EveEntityFactory):
    name = factory.Sequence(lambda n: f"character_name_{n}")
    category = EveEntity.CATEGORY_CHARACTER


class EveEntityCorporationFactory(EveEntityFactory):
    name = factory.Sequence(lambda n: f"corporation_name_{n}")
    category = EveEntity.CATEGORY_CORPORATION


class EveEntityAllianceFactory(EveEntityFactory):
    name = factory.Sequence(lambda n: f"alliance_name_{n}")
    category = EveEntity.CATEGORY_ALLIANCE


# Structures objects


class SuperuserFactory(UserFactory):
    is_staff = True
    is_superuser = True


class UserMainBasicFactory(UserMainFactory):
    """Basic user in Structures."""

    main_character__scopes = Owner.get_esi_scopes()
    permissions__ = ["structures.basic_access"]


class UserMainDefaultFactory(UserMainFactory):
    """Default user in Structures."""

    main_character__scopes = Owner.get_esi_scopes()
    permissions__ = [
        "structures.basic_access",
        "structures.view_corporation_structures",
    ]


class UserMainDefaultOwnerFactory(UserMainFactory):
    """Default user owning structures."""

    main_character__scopes = Owner.get_esi_scopes()
    permissions__ = [
        "structures.basic_access",
        "structures.add_structure_owner",
        "structures.view_corporation_structures",
    ]


class FuelAlertConfigFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[FuelAlertConfig]
):
    class Meta:
        model = FuelAlertConfig

    start = 48
    end = 0
    repeat = 12


class FuelAlertFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[FuelAlert]
):
    class Meta:
        model = FuelAlert

    config = factory.SubFactory(FuelAlertConfigFactory)
    hours = 12


class JumpFuelAlertConfigFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[JumpFuelAlertConfig]
):
    class Meta:
        model = JumpFuelAlertConfig

    threshold = 100


class WebhookFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Webhook]
):
    class Meta:
        model = Webhook
        django_get_or_create = ("name",)

    name = factory.Sequence(lambda n: f"Generated webhook #{n+1}")
    url = factory.LazyAttribute(lambda o: f"http://www.example.com/{slugify(o.name)}")
    notes = factory.Faker("sentence")


class OwnerFactory(factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Owner]):
    class Meta:
        model = Owner
        django_get_or_create = ("corporation",)

    class Params:
        user = None  # when specified will use corporation of users's main

    assets_last_update_at = factory.LazyFunction(now)
    character_ownership = None  # no longer used
    forwarding_last_update_at = factory.LazyFunction(now)
    is_alliance_main = False
    is_up = True
    notifications_last_update_at = factory.LazyFunction(now)
    structures_last_update_at = factory.LazyFunction(now)

    @factory.lazy_attribute
    def corporation(self):
        if self.user:
            corporation = self.user.profile.main_character.corporation
            if corporation:
                return corporation
        return EveCorporationInfoFactory()

    @factory.post_generation
    def characters(
        obj, create: bool, extracted: Optional[List[EveCharacter]], **kwargs
    ):
        # Set characters=False to skip creating characters.
        if not create or extracted is False:
            return

        if extracted:
            for eve_character in extracted:
                character_ownership = eve_character.character_ownership
                OwnerCharacterFactory(
                    owner=obj, character_ownership=character_ownership, **kwargs
                )
            return

        # generate new random owner character from this corporation
        OwnerCharacterFactory(owner=obj, **kwargs)

    @factory.post_generation
    def webhooks(obj, create, extracted, **kwargs):
        # Set webhooks=False to skip creating characters.
        if not create or extracted is False:
            return

        if extracted:
            for webhook in extracted:
                obj.webhooks.add(webhook)

        else:
            obj.webhooks.add(WebhookFactory(**kwargs))


class OwnerCharacterFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[OwnerCharacter]
):
    class Meta:
        model = OwnerCharacter

    class Params:
        eve_character = None

    structures_last_used_at = factory.LazyFunction(now)
    notifications_last_used_at = factory.LazyFunction(now)

    @factory.lazy_attribute
    def character_ownership(self):
        if not self.eve_character:
            character = EveCharacterFactory(corporation=self.owner.corporation)
        else:
            character = self.eve_character
        user = UserMainDefaultOwnerFactory(main_character__character=character)
        return user.profile.main_character.character_ownership


class StructureFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Structure]
):
    class Meta:
        model = Structure
        django_get_or_create = ("id",)

    class Params:
        eve_type_name = "Astrahus"
        eve_solar_system_name = "Amamake"

    id = factory.Sequence(lambda n: 1_500_000_000_000 + n)
    fuel_expires_at = factory.LazyAttribute(lambda obj: now() + dt.timedelta(days=3))
    has_fitting = False
    has_core = True
    last_updated_at = factory.LazyFunction(now)
    name = factory.LazyAttribute(lambda o: f"Test Structure #{o.id}")
    owner = factory.SubFactory(OwnerFactory)
    position_x = factory.fuzzy.FuzzyFloat(-10_000_000_000_000, 10_000_000_000_000)
    position_y = factory.fuzzy.FuzzyFloat(-10_000_000_000_000, 10_000_000_000_000)
    position_z = factory.fuzzy.FuzzyFloat(-10_000_000_000_000, 10_000_000_000_000)
    state = Structure.State.SHIELD_VULNERABLE

    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(name=self.eve_type_name)

    @factory.lazy_attribute
    def eve_solar_system(self):
        return EveSolarSystem.objects.get(name=self.eve_solar_system_name)

    @factory.post_generation
    def webhooks(obj, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for webhook in extracted:
                obj.webhooks.add(webhook)

    @factory.post_generation
    def tags(obj, create, extracted, **kwargs):
        if not create:
            return

        elif extracted:
            obj.tags.add(*extracted)

    @factory.post_generation
    def quantum_core(obj, create, extracted, **kwargs):
        if not create or extracted is False:
            return

        if obj.eve_type.eve_group_id not in {
            EveGroupId.CITADEL,
            EveGroupId.ENGINEERING_COMPLEX,
            EveGroupId.REFINERY,
        }:
            return

        # add quantum core
        StructureItemFactory(
            structure=obj,
            location_flag=StructureItem.LocationFlag.QUANTUM_CORE_ROOM,
            eve_type=EveType.objects.get(
                name=f"{obj.eve_type.name} Upwell Quantum Core"
            ),
        )


class RefineryFactory(StructureFactory):
    class Params:
        eve_moon_name = "Amamake IV - Moon 1"

    @factory.lazy_attribute
    def eve_moon(self):
        return EveMoon.objects.get(name=self.eve_moon_name)

    @factory.lazy_attribute
    def eve_solar_system(self):
        return self.eve_moon.eve_planet.eve_solar_system

    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(name="Athanor")


class StarbaseFactory(StructureFactory):
    class Params:
        eve_moon_name = "Amamake II - Moon 1"

    has_fitting = None
    has_core = None
    state = Structure.State.POS_ONLINE

    @factory.lazy_attribute
    def eve_moon(self):
        return EveMoon.objects.get(name=self.eve_moon_name)

    @factory.lazy_attribute
    def eve_solar_system(self):
        return self.eve_moon.eve_planet.eve_solar_system

    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(name="Caldari Control Tower")

    @factory.post_generation
    def starbase_detail(obj, create, extracted, **kwargs):
        """Set this param to False to disable.

        Set StarbaseDetails attributes with `starbase_detail__key=value`
        """
        if not create or extracted is False:
            return

        StarbaseDetailFactory(structure=obj, **kwargs)


class StarbaseDetailFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[StarbaseDetail]
):
    class Meta:
        model = StarbaseDetail

    structure = factory.SubFactory(StarbaseFactory, starbase_detail=False)

    allow_alliance_members = False
    allow_corporation_members = False
    anchor_role = StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE
    attack_if_at_war = False
    attack_if_other_security_status_dropping = False
    fuel_bay_take_role = StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE
    fuel_bay_view_role = StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE
    last_modified_at = factory.LazyFunction(now)
    offline_role = StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE
    online_role = StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE
    unanchor_role = StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE
    use_alliance_standings = False

    @factory.post_generation
    def fuel_detail(obj, create, extracted, **kwargs):
        """Set this param to False to disable."""
        if not create or extracted is False:
            return

        StarbaseDetailFuelFactory(
            detail=obj, eve_type_name="Nitrogen Fuel Block", quantity=960
        )
        StarbaseDetailFuelFactory(
            detail=obj, eve_type_name="Strontium Clathrates", quantity=12_000
        )


class StarbaseDetailFuelFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[StarbaseDetailFuel]
):
    class Meta:
        model = StarbaseDetailFuel

    class Params:
        eve_type_name = "Nitrogen Fuel Block"

    quantity = 1000

    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(name=self.eve_type_name)


class PocoFactory(StructureFactory):
    class Params:
        eve_planet_name = "Amamake V"

    has_fitting = None
    has_core = None
    state = Structure.State.NA

    @factory.lazy_attribute
    def eve_planet(self):
        return EvePlanet.objects.get(name=self.eve_planet_name)

    @factory.lazy_attribute
    def eve_solar_system(self):
        return self.eve_planet.eve_solar_system

    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(name="Customs Office")

    @factory.lazy_attribute
    def name(self):
        return f"Customs Office ({self.eve_planet.name})"

    @factory.post_generation
    def poco_details(obj, create, extracted, **kwargs):
        """Set this param to False to disable.

        Set PocoDetails attributes with `poco_details__key=value`
        """
        if not create or extracted is False:
            return

        PocoDetailsFactory(structure=obj, **kwargs)


class PocoDetailsFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[PocoDetails]
):
    class Meta:
        model = PocoDetails

    structure = factory.SubFactory(PocoFactory, poco_details=False)

    allow_access_with_standings = False
    allow_alliance_access = False
    reinforce_exit_end = 21
    reinforce_exit_start = 18

    standing_level = PocoDetails.StandingLevel.BAD


class SkyhookFactory(StructureFactory):
    class Params:
        eve_planet_name = "Amamake V"

    has_fitting = None
    has_core = None
    state = Structure.State.NA

    @factory.lazy_attribute
    def eve_planet(self):
        return EvePlanet.objects.get(name=self.eve_planet_name)

    @factory.lazy_attribute
    def eve_solar_system(self):
        return self.eve_planet.eve_solar_system

    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(id=EveTypeId.ORBITAL_SKYHOOK)


class JumpGateFactory(StructureFactory):
    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(name="Ansiblex Jump Gate")

    @factory.post_generation
    def jump_fuel_quantity(obj, create, extracted, **kwargs):
        """Set this param to False to disable."""
        if not create or extracted is False:
            return

        StructureItemJumpFuelFactory(
            structure=obj,
            quantity=extracted or 1000,  # default
        )


class StructureItemFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[StructureItem]
):
    class Meta:
        model = StructureItem
        django_get_or_create = ("id",)

    class Params:
        eve_type_name = "Oxygen Fuel Block"

    id = factory.Sequence(lambda n: 1900000000001 + n)
    is_singleton = False
    location_flag = StructureItem.LocationFlag.CARGO
    quantity = 1

    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(name=self.eve_type_name)


class StructureItemJumpFuelFactory(StructureItemFactory):
    is_singleton = False
    location_flag = StructureItem.LocationFlag.STRUCTURE_FUEL

    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(id=EveTypeId.LIQUID_OZONE)


class StructureServiceFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[StructureService]
):
    class Meta:
        model = StructureService
        django_get_or_create = ("name",)

    name = factory.Sequence(lambda n: f"Fake Service #{n + 1}")
    state = StructureService.State.ONLINE


class StructureTagFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[StructureTag]
):
    class Meta:
        model = StructureTag
        django_get_or_create = ("name",)

    name = factory.Sequence(lambda n: f"name_{n}")
    description = factory.Faker("sentence")


class EveSovereigntyMapFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[EveSovereigntyMap]
):
    class Meta:
        model = EveSovereigntyMap

    class Params:
        corporation = None
        eve_solar_system_name = "1-PGSG"

    last_updated = factory.LazyFunction(now)

    @factory.lazy_attribute
    def solar_system_id(self):
        obj = EveSolarSystem.objects.get(name=self.eve_solar_system_name)
        return obj.id

    @factory.lazy_attribute
    def alliance_id(self):
        if self.corporation:
            if self.corporation.alliance:
                return self.corporation.alliance.alliance_id
            raise ValueError("Corporation must be in alliance to have sov")
        return None

    @factory.lazy_attribute
    def corporation_id(self):
        if self.corporation:
            return self.corporation.corporation_id
        return None


class NotificationFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Notification]
):
    class Meta:
        model = Notification

    class Params:
        text_from_dict = None

    notification_id = factory.Sequence(lambda n: 1_500_000_000 + n)
    created = factory.LazyFunction(now)
    is_read = False
    last_updated = factory.LazyAttribute(lambda o: o.created)
    notif_type = NotificationType.WAR_CORPORATION_BECAME_ELIGIBLE
    owner = factory.SubFactory(OwnerFactory)
    sender = factory.SubFactory(EveEntityCorporationFactory, id=1000137, name="DED")
    timestamp = factory.LazyAttribute(lambda o: o.created)

    @factory.lazy_attribute
    def text(self):
        if not self.text_from_dict:
            return ""
        return yaml.dump(self.text_from_dict)

    @classmethod
    def _adjust_kwargs(cls, **kwargs):
        if isinstance(kwargs["notif_type"], NotificationType):
            kwargs["notif_type"] = kwargs["notif_type"].value
        return kwargs


class NotificationMoonMiningExtractionStartedFactory(NotificationFactory):
    class Params:
        auto_time = 132186924601059151
        ore_volume_by_type = {
            46300: 1288475.124715103,
            46301: 544691.7637724016,
            46302: 526825.4047522942,
            46303: 528996.6386983792,
        }
        ready_time = 132186816601059151
        structure = None
        started_by = None

    notif_type = NotificationType.MOONMINING_EXTRACTION_STARTED

    @factory.lazy_attribute
    def text(self):
        started_by = self.started_by or EveEntityCharacterFactory()
        structure = self.structure or RefineryFactory(owner=self.owner)
        data = {
            "autoTime": self.auto_time,
            "moonID": structure.eve_moon.id,
            "oreVolumeByType": self.ore_volume_by_type,
            "readyTime": self.ready_time,
            "solarSystemID": structure.eve_solar_system.id,
            "startedBy": started_by.id,
            "startedByLink": (
                f'<a href="showinfo:1383\\/\\/{started_by.id}">'
                f"{started_by.name}<\\/a>"
            ),
            "structureID": structure.id,
            "structureLink": (
                f'<a href="showinfo:{structure.eve_type.id}\\/\\/'
                f'{structure.id}">{structure.name}<\\/a>'
            ),
            "structureName": structure.name,
            "structureTypeID": structure.eve_type.id,
        }
        return yaml.dump(data)


class NotificationMoonMiningExtractionCanceledFactory(NotificationFactory):
    class Params:
        canceled_by = None
        structure = None

    notif_type = NotificationType.MOONMINING_EXTRACTION_CANCELLED

    @factory.lazy_attribute
    def text(self):
        canceled_by = self.canceled_by or EveEntityCharacterFactory()
        structure = self.structure or RefineryFactory(owner=self.owner)

        data = {
            "cancelledBy": canceled_by.id,
            "cancelledByLink": (
                f'<a href="showinfo:1383\\/\\/{canceled_by.id}">'
                f"{canceled_by.name}<\\/a>"
            ),
            "moonID": structure.eve_moon.id,
            "solarSystemID": structure.eve_solar_system.id,
            "structureID": structure.id,
            "structureLink": (
                f'<a href="showinfo:{structure.eve_type.id}\\/\\/'
                f'{structure.id}">{structure.name}<\\/a>'
            ),
            "structureName": structure.name,
            "structureTypeID": structure.eve_type.id,
        }
        return yaml.dump(data)


class NotificationOrbitalReinforcedFactory(NotificationFactory):
    class Params:
        aggressor_alliance = None
        aggressor_corporation = None
        aggressor_character = None
        structure = None

    notif_type = NotificationType.ORBITAL_REINFORCED

    @factory.lazy_attribute
    def text(self):
        structure = self.structure or PocoFactory(owner=self.owner)
        aggressor_alliance = self.aggressor_alliance or EveEntityAllianceFactory()
        aggressor_corporation = (
            self.aggressor_corporation or EveEntityCorporationFactory()
        )
        aggressor_character = self.aggressor_character or EveEntityCharacterFactory()
        data = {
            "aggressorAllianceID": aggressor_alliance.id,
            "aggressorCorpID": aggressor_corporation.id,
            "aggressorID": aggressor_character.id,
            "planetID": structure.eve_planet.id,
            "planetTypeID": structure.eve_planet.eve_type.id,
            "reinforceExitTime": 132154723470000000,  # TODO: make variable
            "solarSystemID": structure.eve_solar_system.id,
            "typeID": structure.eve_type.id,
        }
        return yaml.dump(data)


class NotificationStructureLostShieldFactory(NotificationFactory):
    class Params:
        structure = None

    notif_type = NotificationType.STRUCTURE_LOST_SHIELD

    @factory.lazy_attribute
    def text(self):
        structure = self.structure or StructureFactory(owner=self.owner)
        data = {
            "solarsystemID": structure.eve_solar_system.id,
            "structureID": structure.id,
            "structureShowInfoData": [
                "showinfo",
                structure.eve_type.id,
                structure.id,
            ],
            "structureTypeID": structure.eve_type.id,
            "timeLeft": 1727805401093,
            "timestamp": 132148470780000000,
            "vulnerableTime": 9000000000,
        }
        return yaml.dump(data)


class NotificationSovStructureReinforcedFactory(NotificationFactory):
    class Params:
        campaign_event_type = 1
        decloak_time = 131897990021334067
        eve_solar_system_name = "1-PGSG"

    notif_type = NotificationType.SOV_STRUCTURE_REINFORCED

    @factory.lazy_attribute
    def text(self):
        eve_solar_system = EveSolarSystem.objects.get(name=self.eve_solar_system_name)
        data = {
            "campaignEventType": self.campaign_event_type,
            "decloakTime": self.decloak_time,
            "solarSystemID": eve_solar_system.id,
        }
        return yaml.dump(data)


class GeneratedNotificationFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[GeneratedNotification]
):
    class Meta:
        model = GeneratedNotification

    notif_type = NotificationType.TOWER_REINFORCED_EXTRA.value
    owner = factory.SubFactory(OwnerFactory)

    @factory.lazy_attribute
    def details(self):
        reinforced_until = factory.fuzzy.FuzzyDateTime(
            start_dt=now() + dt.timedelta(hours=3),
            end_dt=now() + dt.timedelta(hours=48),
        ).fuzz()
        return {"reinforced_until": reinforced_until.isoformat()}

    @factory.post_generation
    def create_structure(obj, create, extracted, **kwargs):
        """Set this param to False to disable."""
        if not create or extracted is False:
            return

        reinforced_until = dt.datetime.fromisoformat(obj.details["reinforced_until"])
        starbase = StarbaseFactory(
            owner=obj.owner,
            state=Structure.State.POS_REINFORCED,
            state_timer_end=reinforced_until,
        )
        obj.structures.add(starbase)


class RawNotificationFactory(factory.DictFactory, metaclass=BaseMetaFactory[dict]):
    """Create a raw notification as received from ESI."""

    class Meta:
        exclude = ("data", "timestamp_dt", "sender")

    # excluded
    data = None
    timestamp_dt = None
    sender = factory.SubFactory(EveEntityCorporationFactory, id=2902, name="CONCORD")

    # included
    notification_id = factory.Sequence(lambda o: 1999000000 + o)
    type = "CorpBecameWarEligible"
    sender_id = factory.LazyAttribute(lambda o: o.sender.id)
    sender_type = factory.LazyAttribute(lambda o: o.sender.category)
    is_read = True

    @factory.lazy_attribute
    def timestamp(self):
        if not self.timestamp_dt:
            timestamp_dt = now()
        else:
            timestamp_dt = self.timestamp_dt
        return datetime_to_esi(timestamp_dt)

    @factory.lazy_attribute
    def text(self):
        if not self.data:
            return ""
        return yaml.dump(self.data)



"""War embeds."""

# pylint: disable=missing-class-docstring


import dhooks_lite

from django.utils.html import strip_tags
from django.utils.translation import gettext as _

from app_utils.datetime import ldap_time_2_datetime
from app_utils.helpers import humanize_number

from structures.helpers import get_or_create_eve_entity
from structures.models import Notification, Webhook

from .helpers import gen_eve_entity_link, target_datetime_formatted
from .main import NotificationBaseEmbed


class NotificationAcceptedAlly(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Accepted Ally")
        ally = get_or_create_eve_entity(id=self._data["allyID"])
        enemy = get_or_create_eve_entity(id=self._data["enemyID"])
        character = get_or_create_eve_entity(id=self._data["charID"])
        isk_value = self._data["iskValue"]
        time = ldap_time_2_datetime(self._data["time"])
        self._description = _(
            "%(ally)s has been accepted in a war against %(enemy)s. "
            "The offer was accepted at %(time)s by %(character)s for %(isk_value)s ISK."
        ) % {
            "enemy": gen_eve_entity_link(enemy),
            "ally": gen_eve_entity_link(ally),
            "character": gen_eve_entity_link(character),
            "isk_value": humanize_number(isk_value),
            "time": target_datetime_formatted(time),
        }
        self._thumbnail = dhooks_lite.Thumbnail(
            ally.icon_url(size=self.ICON_DEFAULT_SIZE)
        )
        self._color = Webhook.Color.WARNING


class NotificationAllWarCorpJoinedAllianceMsg(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Corporation you are at war with is joining an alliance")
        alliance = get_or_create_eve_entity(id=self._data["allianceID"])
        corporation = get_or_create_eve_entity(id=self._data["corpID"])
        self._description = _(
            "%(corporation)s is joining %(alliance)s alliance. "
            "Since you are at war with %(corporation)s, "
            "in 24 hours you will also be at war with %(alliance)s."
        ) % {
            "alliance": gen_eve_entity_link(alliance),
            "corporation": gen_eve_entity_link(corporation),
        }
        self._thumbnail = dhooks_lite.Thumbnail(
            corporation.icon_url(size=self.ICON_DEFAULT_SIZE)
        )
        self._color = Webhook.Color.INFO


class NotificationAllyJoinedWarMsg(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Ally Has Joined a War")
        aggressor = get_or_create_eve_entity(id=self._data["aggressorID"])
        ally = get_or_create_eve_entity(id=self._data["allyID"])
        defender = get_or_create_eve_entity(id=self._data["defenderID"])
        start_time = ldap_time_2_datetime(self._data["startTime"])
        self._description = _(
            "%(ally)s has joined %(defender)s in a war against %(aggressor)s. "
            "Their participation in the war will start at %(start_time)s."
        ) % {
            "aggressor": gen_eve_entity_link(aggressor),
            "ally": gen_eve_entity_link(ally),
            "defender": gen_eve_entity_link(defender),
            "start_time": target_datetime_formatted(start_time),
        }
        self._thumbnail = dhooks_lite.Thumbnail(
            ally.icon_url(size=self.ICON_DEFAULT_SIZE)
        )
        self._color = Webhook.Color.WARNING


class NotificationDeclareWar(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        entity = get_or_create_eve_entity(id=self._data["entityID"])
        defender = get_or_create_eve_entity(id=self._data["defenderID"])
        self._title = _("%(entity)s Declares War Against %(defender)s") % {
            "entity": entity.name,
            "defender": defender.name,
        }
        self._description = _(
            "%(entity)s has declared war on %(defender)s. "
            "Within 24 hours fighting can legally occur between those involved."
        ) % {
            "entity": gen_eve_entity_link(entity),
            "defender": gen_eve_entity_link(defender),
        }
        self._thumbnail = dhooks_lite.Thumbnail(
            entity.icon_url(size=self.ICON_DEFAULT_SIZE)
        )
        self._color = Webhook.Color.DANGER


class NotificationMercOfferedNegotiationMsg(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        aggressor = get_or_create_eve_entity(id=self._data["aggressorID"])
        defender = get_or_create_eve_entity(id=self._data["defenderID"])
        mercenary = get_or_create_eve_entity(id=self._data["mercID"])
        isk_value = self._data["iskValue"]
        self._title = (
            _("%s has offered its services in one of your wars") % mercenary.name
        )
        self._description = _(
            "%(mercenary)s has offered %(defender)s it's services in it's war against %(aggressor)s for %(isk_value)s. "
        ) % {
            "aggressor": gen_eve_entity_link(aggressor),
            "defender": gen_eve_entity_link(defender),
            "mercenary": gen_eve_entity_link(mercenary),
            "isk_value": humanize_number(isk_value),
        }
        self._thumbnail = dhooks_lite.Thumbnail(
            mercenary.icon_url(size=self.ICON_DEFAULT_SIZE)
        )
        self._color = Webhook.Color.INFO


class NotificationMercOfferRetractedMsg(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Mercenary offered services")
        aggressor = get_or_create_eve_entity(id=self._data["aggressorID"])
        defender = get_or_create_eve_entity(id=self._data["defenderID"])
        mercenary = get_or_create_eve_entity(id=self._data["mercID"])
        self._description = _(
            "%(mercenary)s has retracted it's offer to support %(defender)s in a war against %(aggressor)s."
        ) % {
            "aggressor": gen_eve_entity_link(aggressor),
            "defender": gen_eve_entity_link(defender),
            "mercenary": gen_eve_entity_link(mercenary),
        }
        self._thumbnail = dhooks_lite.Thumbnail(
            mercenary.icon_url(size=self.ICON_DEFAULT_SIZE)
        )
        self._color = Webhook.Color.INFO


class NotificationOfferedToAlly(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        aggressor = get_or_create_eve_entity(id=self._data["aggressorID"])
        defender = get_or_create_eve_entity(id=self._data["defenderID"])
        character = get_or_create_eve_entity(id=self._data["mercID"])
        isk_value = self._data["iskValue"]
        self._title = _("You have offered to ally with %s ") % defender.name
        self._description = _(
            "%(character)s has offered to ally to %(defender)s in their war "
            "against %(aggressor)s. The offer asked for %(isk_value)s ISK as payment."
        ) % {
            "aggressor": gen_eve_entity_link(aggressor),
            "defender": gen_eve_entity_link(defender),
            "character": gen_eve_entity_link(character),
            "isk_value": humanize_number(isk_value),
        }
        self._thumbnail = dhooks_lite.Thumbnail(
            character.icon_url(size=self.ICON_DEFAULT_SIZE)
        )
        self._color = Webhook.Color.INFO


class NotificationOfferedSurrender(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        character = get_or_create_eve_entity(id=self._data["charID"])
        entity = get_or_create_eve_entity(id=self._data["entityID"])
        offered = get_or_create_eve_entity(id=self._data["offeredID"])
        isk_value = self._data["iskValue"]
        self._title = _("%s offered to surrender") % entity.name
        self._description = _(
            "%(character)s has offered to surrender to %(offered)s, "
            "offering %(isk_value)s ISK. If accepted, the war will end in 24 hours "
            "and neither organization will be able to declare new wars against "
            "the other for the next 2 weeks."
        ) % {
            "character": gen_eve_entity_link(character),
            "offered": gen_eve_entity_link(offered),
            "isk_value": humanize_number(isk_value),
        }
        self._color = Webhook.Color.WARNING


class NotificationWarCorporationBecameEligible(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _(
            "Corporation or alliance is now eligible for formal war declarations"
        )
        self._description = _(
            "Your corporation or alliance is **now eligible** to participate in "
            "formal war declarations. This could be because your corporation "
            "and/or one of the corporations in your alliance owns a structure "
            "deployed in space."
        )
        self._color = Webhook.Color.WARNING


class NotificationWarCorporationNoLongerEligible(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _(
            "Corporation or alliance is no longer eligible for formal war declarations"
        )
        self._description = _(
            "Your corporation or alliance is **no longer eligible** to participate "
            "in formal war declarations.\n"
            "Neither your corporation nor any of the corporations "
            "in your alliance own a structure deployed in space at this time. "
            "If your corporation or alliance is currently involved in a formal war, "
            "that war will end in 24 hours."
        )
        self._color = Webhook.Color.INFO


class NotificationWarSurrenderOfferMsg(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        owner_1 = get_or_create_eve_entity(id=self._data.get("ownerID1"))
        owner_2 = get_or_create_eve_entity(id=self._data.get("ownerID2"))
        isk_value = self._data.get("iskValue", 0)
        self._title = _("%s has offered a surrender") % owner_1.name
        self._description = _(
            "%(owner_1)s has offered to end the war with %(owner_2)s in the exchange "
            "for %(isk_value)s ISK. "
            "If accepted, the war will end in 24 hours and your organizations will "
            "be unable to declare new wars against each other for the next 2 weeks."
        ) % {
            "owner_1": gen_eve_entity_link(owner_1),
            "owner_2": gen_eve_entity_link(owner_2),
            "isk_value": humanize_number(isk_value),
        }
        self._color = Webhook.Color.INFO


#################################
# War notifs with common elements
#################################


class NotificationWarBaseEmbed(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._declared_by = get_or_create_eve_entity(id=self._data["declaredByID"])
        self._against = get_or_create_eve_entity(id=self._data["againstID"])
        self._thumbnail = dhooks_lite.Thumbnail(
            self._declared_by.icon_url(size=self.ICON_DEFAULT_SIZE)
        )


class NotificationAllWarSurrenderMsg(NotificationWarBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("%s has surrendered") % self._declared_by.name
        self._description = _(
            "%(declared_by)s has surrendered in the war against  %(against)s."
        ) % {
            "declared_by": gen_eve_entity_link(self._declared_by),
            "against": gen_eve_entity_link(self._against),
        }
        self._color = Webhook.Color.WARNING


class NotificationCorpWarSurrenderMsg(NotificationWarBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("One party has surrendered")
        self._description = _(
            "The war between %(against)s and %(declared_by)s is coming to an end "
            "as one party has surrendered. "
            "The war will be declared as being over after approximately 24 hours."
        ) % {
            "declared_by": gen_eve_entity_link(self._declared_by),
            "against": gen_eve_entity_link(self._against),
        }
        self._color = Webhook.Color.WARNING


class NotificationWarHQRemovedFromSpace(NotificationWarBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        war_hq = self._data["warHQ"]
        time_declared = ldap_time_2_datetime(self._data["timeDeclared"])
        self._title = _("WarHQ %s lost") % war_hq
        self._description = _(
            "The war HQ %(war_hq)s is no more. "
            "As a consequence, the war declared by %(declared_by)s "
            "against %(against)s on %(time_declared)s "
            "has been declared invalid by CONCORD and has entered its cooldown period."
        ) % {
            "war_hq": war_hq,
            "declared_by": self._declared_by,
            "against": self._against,
            "time_declared": time_declared,
        }
        self._color = Webhook.Color.WARNING


class NotificationWarAdopted(NotificationWarBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        alliance = get_or_create_eve_entity(id=self._data["allianceID"])
        self._title = _("War update: %(against)s has left %(alliance)s") % {
            "against": self._against.name,
            "alliance": alliance.name,
        }
        self._description = _(
            "There has been a development in the war between %(declared_by)s "
            "and %(alliance)s.\n"
            "%(against)s is no longer a member of %(alliance)s, "
            "and therefore a new war between %(declared_by)s and %(against)s has begun."
        ) % {
            "declared_by": gen_eve_entity_link(self._declared_by),
            "against": gen_eve_entity_link(self._against),
            "alliance": gen_eve_entity_link(alliance),
        }
        self._color = Webhook.Color.WARNING


class NotificationWarDeclared(NotificationWarBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("%(declared_by)s Declares War Against %(against)s") % {
            "declared_by": self._declared_by.name,
            "against": self._against.name,
        }
        self._description = _(
            "%(declared_by)s has declared war on %(against)s with %(war_hq)s "
            "as the designated war headquarters.\n"
            "Within %(delay_hours)s hours fighting can legally occur "
            "between those involved."
        ) % {
            "declared_by": gen_eve_entity_link(self._declared_by),
            "against": gen_eve_entity_link(self._against),
            "war_hq": Webhook.text_bold(strip_tags(self._data["warHQ"])),
            "delay_hours": Webhook.text_bold(self._data["delayHours"]),
        }
        self._color = Webhook.Color.DANGER


class NotificationWarInherited(NotificationWarBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        alliance = get_or_create_eve_entity(id=self._data["allianceID"])
        opponent = get_or_create_eve_entity(id=self._data["opponentID"])
        quitter = get_or_create_eve_entity(id=self._data["quitterID"])
        self._title = _("%(alliance)s inherits war against %(opponent)s") % {
            "alliance": alliance.name,
            "opponent": opponent.name,
        }
        self._description = _(
            "%(alliance)s has inherited the war between %(declared_by)s and "
            "%(against)s from newly joined %(quitter)s. "
            "Within **24** hours fighting can legally occur with %(alliance)s."
        ) % {
            "declared_by": gen_eve_entity_link(self._declared_by),
            "against": gen_eve_entity_link(self._against),
            "alliance": gen_eve_entity_link(alliance),
            "quitter": gen_eve_entity_link(quitter),
        }
        self._color = Webhook.Color.DANGER


class NotificationWarInvalid(NotificationWarBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        war_ends = ldap_time_2_datetime(self._data["endDate"])
        self._title = _("CONCORD invalidates war")
        self._description = _(
            "The war between %(declared_by)s and %(against)s "
            "has been retracted by CONCORD, "
            "because at least one of the involved parties "
            "has become ineligible for war declarations."
            "Fighting must cease on %(end_date)s."
        ) % {
            "declared_by": gen_eve_entity_link(self._declared_by),
            "against": gen_eve_entity_link(self._against),
            "end_date": target_datetime_formatted(war_ends),
        }
        self._color = Webhook.Color.WARNING


class NotificationWarRetractedByConcord(NotificationWarBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("CONCORD retracts war")
        war_ends = ldap_time_2_datetime(self._data["endDate"])
        self._description = _(
            "The war between %(declared_by)s and %(against)s "
            "has been retracted by CONCORD.\n"
            "After %(end_date)s CONCORD will again respond to any hostilities "
            "between those involved with full force."
        ) % {
            "declared_by": gen_eve_entity_link(self._declared_by),
            "against": gen_eve_entity_link(self._against),
            "end_date": target_datetime_formatted(war_ends),
        }
        self._color = Webhook.Color.WARNING



from unittest.mock import patch

from django.test import TestCase
from django.urls import reverse

from app_utils.testing import add_character_to_user

from structures.tests.testdata.factories import (
    EveCharacterFactory,
    JumpGateFactory,
    OwnerFactory,
    PocoFactory,
    StarbaseFactory,
    StructureFactory,
    UserMainBasicFactory,
    UserMainDefaultFactory,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

STRUCTURES_PATH = "structures.views.structures"
PUBLIC_PATH = "structures.views.public"
STATISTICS_PATH = "structures.views.statistics"
COMMON_PATH = "structures.views.common"


class TestStructureListView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.user = UserMainDefaultFactory()
        cls.owner = OwnerFactory(user=cls.user)

    @patch(COMMON_PATH + ".STRUCTURES_DEFAULT_PAGE_LENGTH", 15)
    @patch(COMMON_PATH + ".STRUCTURES_PAGING_ENABLED", True)
    def test_should_return_correct_context(self):
        # given
        StructureFactory(owner=self.owner)
        PocoFactory(owner=self.owner)
        StarbaseFactory(owner=self.owner)
        JumpGateFactory(owner=self.owner)
        StructureFactory()  # this one will be hidden
        self.client.force_login(self.user)
        # when
        response = self.client.get(reverse("structures:structure_list"))
        # then
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context["structures_count"], 2)
        self.assertEqual(response.context["orbitals_count"], 1)
        self.assertEqual(response.context["starbases_count"], 1)
        self.assertEqual(response.context["jump_gates_count"], 1)
        self.assertIn("data_export", response.context)
        data_export = response.context["data_export"]
        self.assertEqual(data_export["data_tables_page_length"], 15)
        self.assertEqual(data_export["data_tables_paging"], 1)


class TestStatisticsView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.user = UserMainDefaultFactory()
        cls.owner = OwnerFactory(user=cls.user)

    @patch(COMMON_PATH + ".STRUCTURES_DEFAULT_PAGE_LENGTH", 15)
    @patch(COMMON_PATH + ".STRUCTURES_PAGING_ENABLED", True)
    def test_should_return_correct_context(self):
        # given
        StructureFactory(owner=self.owner)
        self.client.force_login(self.user)
        # when
        response = self.client.get(reverse("structures:statistics"))
        # then
        self.assertEqual(response.status_code, 200)
        self.assertIn("last_updated", response.context)
        self.assertIn("data_export", response.context)
        data_export = response.context["data_export"]
        self.assertEqual(data_export["data_tables_page_length"], 15)
        self.assertEqual(data_export["data_tables_paging"], 1)


class TestPocoView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        character = EveCharacterFactory(character_name="Bruce Wayne")
        cls.user = UserMainBasicFactory(main_character__character=character)
        cls.owner = OwnerFactory(are_pocos_public=True)
        cls.alt_character = EveCharacterFactory(character_name="Peter Parker")
        add_character_to_user(cls.user, cls.alt_character)

    @patch(COMMON_PATH + ".STRUCTURES_DEFAULT_PAGE_LENGTH", 15)
    @patch(COMMON_PATH + ".STRUCTURES_PAGING_ENABLED", True)
    def test_should_return_correct_context(self):
        # given
        PocoFactory(owner=self.owner)
        PocoFactory(owner=self.owner)
        PocoFactory()  # this one will be hidden
        self.client.force_login(self.user)
        # when
        response = self.client.get(reverse("structures:public"))
        # then
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context["pocos_count"], 2)
        characters = {obj["character_name"] for obj in response.context["characters"]}
        self.assertEqual(characters, {"Bruce Wayne", "Peter Parker"})
        self.assertIn("data_export", response.context)
        data_export = response.context["data_export"]
        self.assertEqual(data_export["data_tables_page_length"], 15)
        self.assertEqual(data_export["data_tables_paging"], 1)



"""Sovereignty embeds."""

# pylint: disable=missing-class-docstring

import dhooks_lite

from django.utils.translation import gettext as _
from eveuniverse.models import EveMoon

from app_utils.datetime import ldap_time_2_datetime

from structures.constants import EveTypeId
from structures.core import sovereignty
from structures.helpers import (
    get_or_create_esi_obj,
    get_or_create_eve_entity,
    get_or_create_eve_type,
)
from structures.models import Notification, Webhook

from .helpers import (
    gen_alliance_link,
    gen_corporation_link,
    gen_eve_entity_link,
    gen_solar_system_text,
    target_datetime_formatted,
)
from .main import NotificationBaseEmbed


class NotificationSovEmbed(NotificationBaseEmbed):
    """Base class for all sovereignty related notification embeds."""

    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._solar_system = self._notification.eve_solar_system()
        self._solar_system_link = gen_solar_system_text(self._solar_system)
        if "structureTypeID" in self._data:
            structure_type_id = self._data["structureTypeID"]
        elif "campaignEventType" in self._data:
            structure_type_id = sovereignty.event_type_to_type_id(
                self._data["campaignEventType"]
            )
        else:
            structure_type_id = EveTypeId.TCU
        structure_type = get_or_create_eve_type(id=structure_type_id)
        self._structure_type_name = structure_type.name
        try:
            self._sov_owner_link = gen_alliance_link(notification.sender.name)
        except AttributeError:
            self._sov_owner_link = "(unknown)"
        self._thumbnail = dhooks_lite.Thumbnail(
            structure_type.icon_url(size=self.ICON_DEFAULT_SIZE)
        )


class NotificationSovEntosisCaptureStarted(NotificationSovEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("%(structure_type)s in %(solar_system)s is being captured") % {
            "structure_type": Webhook.text_bold(self._structure_type_name),
            "solar_system": self._solar_system.name,
        }
        self._description = _(
            "A capsuleer has started to influence the %(type)s "
            "in %(solar_system)s belonging to %(owner)s "
            "with an Entosis Link."
        ) % {
            "type": self._structure_type_name,
            "solar_system": self._solar_system_link,
            "owner": self._sov_owner_link,
        }
        self._color = Webhook.Color.WARNING


class NotificationSovCommandNodeEventStarted(NotificationSovEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _(
            "Command nodes for %(structure_type)s in %(solar_system)s "
            "have begun to decloak"
        ) % {
            "structure_type": Webhook.text_bold(self._structure_type_name),
            "solar_system": self._solar_system.name,
        }
        self._description = _(
            "Command nodes for %(structure_type)s in %(solar_system)s "
            "belonging to %(owner)s can now be found throughout "
            "the %(constellation)s constellation"
        ) % {
            "structure_type": Webhook.text_bold(self._structure_type_name),
            "solar_system": self._solar_system_link,
            "owner": self._sov_owner_link,
            "constellation": self._solar_system.eve_constellation.name,
        }
        self._color = Webhook.Color.WARNING


class NotificationSovAllClaimAcquiredMsg(NotificationSovEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        alliance = get_or_create_eve_entity(id=self._data["allianceID"])
        corporation = get_or_create_eve_entity(id=self._data["corpID"])
        self._title = (
            _("DED Sovereignty claim acknowledgment: %s") % self._solar_system.name
        )
        self._description = _(
            "DED now officially acknowledges that your "
            "member corporation %(corporation)s has claimed "
            "sovereignty on behalf of %(alliance)s in %(solar_system)s."
        ) % {
            "corporation": gen_corporation_link(corporation.name),
            "alliance": gen_alliance_link(alliance.name),
            "solar_system": self._solar_system_link,
        }
        self._color = Webhook.Color.SUCCESS


class NotificationSovAllClaimLostMsg(NotificationSovEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        alliance = get_or_create_eve_entity(id=self._data["allianceID"])
        corporation = get_or_create_eve_entity(id=self._data["corpID"])
        self._title = _("Lost sovereignty in: %s") % self._solar_system.name
        self._description = _(
            "DED acknowledges that member corporation %(corporation)s has lost its "
            "claim to sovereignty on behalf of %(alliance)s in %(solar_system)s."
        ) % {
            "corporation": gen_corporation_link(corporation.name),
            "alliance": gen_alliance_link(alliance.name),
            "solar_system": self._solar_system_link,
        }
        self._color = Webhook.Color.SUCCESS


class NotificationSovStructureReinforced(NotificationSovEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        timer_starts = ldap_time_2_datetime(self._data["decloakTime"])
        self._title = _(
            "%(structure_type)s in %(solar_system)s has entered reinforced mode"
        ) % {
            "structure_type": Webhook.text_bold(self._structure_type_name),
            "solar_system": self._solar_system.name,
        }
        self._description = _(
            "The %(structure_type)s in %(solar_system)s belonging "
            "to %(owner)s has been reinforced by "
            "hostile forces and command nodes "
            "will begin decloaking at %(date)s"
        ) % {
            "structure_type": Webhook.text_bold(self._structure_type_name),
            "solar_system": self._solar_system_link,
            "owner": self._sov_owner_link,
            "date": target_datetime_formatted(timer_starts),
        }
        self._color = Webhook.Color.DANGER


class NotificationSovStructureDestroyed(NotificationSovEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("%(structure_type)s in %(solar_system)s has been destroyed") % {
            "structure_type": Webhook.text_bold(self._structure_type_name),
            "solar_system": self._solar_system.name,
        }
        self._description = _(
            "The command nodes for %(structure_type)s "
            "in %(solar_system)s belonging to %(owner)s have been "
            "destroyed by hostile forces."
        ) % {
            "structure_type": Webhook.text_bold(self._structure_type_name),
            "solar_system": self._solar_system_link,
            "owner": self._sov_owner_link,
        }
        self._color = Webhook.Color.DANGER


class NotificationSovAllAnchoringMsg(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        corporation = get_or_create_eve_entity(id=self._data.get("corpID"))
        corp_link = gen_eve_entity_link(corporation)
        alliance_id = self._data.get("allianceID")
        if alliance_id:
            alliance = get_or_create_eve_entity(id=alliance_id)
            structure_owner = f"{corp_link} ({alliance.name})"
        else:
            structure_owner = corp_link
        eve_solar_system = self._notification.eve_solar_system()
        structure_type = self._notification.eve_structure_type("typeID")
        moon_id = self._data.get("moonID")
        if moon_id:
            eve_moon: EveMoon = get_or_create_esi_obj(EveMoon, id=moon_id)
            location_text = _(" near **%s**") % eve_moon.name
        else:
            location_text = ""
        self._title = _("%(structure_type)s anchored in %(solar_system)s") % {
            "structure_type": structure_type.eve_group.name,
            "solar_system": eve_solar_system.name,
        }
        self._description = _(
            "A %(structure_type)s from %(structure_owner)s has anchored "
            "in %(solar_system)s%(location_text)s."
        ) % {
            "structure_type": Webhook.text_bold(structure_type.name),
            "structure_owner": structure_owner,
            "solar_system": gen_solar_system_text(eve_solar_system),
            "location_text": location_text,
        }
        self._color = Webhook.Color.WARNING
        self._thumbnail = dhooks_lite.Thumbnail(
            structure_type.icon_url(size=self.ICON_DEFAULT_SIZE)
        )
